/*
 * Router rc control script
 *
 * Copyright (C) 2009, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: rc.c,v 1.157 2010/02/11 15:26:02 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>
#include <sys/klog.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <sys/reboot.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <net/if_arp.h>
#include <dirent.h>

#include <epivers.h>
#include <router_version.h>
#include <bcmnvram.h>
#include <mtd.h>
#include <shutils.h>
#include <rc.h>
#include <netconf.h>
#include <nvparse.h>
#include <bcmdevs.h>
#include <bcmparams.h>
#include <wlutils.h>
#include <ezc.h>
#if defined(__CONFIG_WAPI__) || defined(__CONFIG_WAPI_IAS__)
#include <wapi_utils.h>
#endif /* __CONFIG_WAPI__ || __CONFIG_WAPI_IAS__ */
#include <fcntl.h>
#include <wps_led.h>

#ifdef __CONFIG_NAT__
static void auto_bridge(void);
#endif	/* __CONFIG_NAT__ */

#include <sys/sysinfo.h> 
#ifdef __CONFIG_EMF__
extern void load_emf(void);
#endif /* __CONFIG_EMF__ */

static void restore_defaults(void);
static void sysinit(void);
static void rc_signal(int sig);

extern struct nvram_tuple router_defaults[];

#define RESTORE_DEFAULTS() \
	(!nvram_match("restore_defaults", "0") || nvram_invmatch("os_name", "linux"))

static int is_wps_sta_mode()
{
    if ( acosNvramConfig_match("wps_success_mode", "0") && 
         acosNvramConfig_match("wps_success_mode_2", WPS_STA_MODE_OK) )
        return 1;
    else
        return 0;
}
static void convert_wlan_params(void)
{
#define WPS_DEBUG 1
        
#ifdef WLAN_REPEATER
        char *wsc_status = acosNvramConfig_get("wps_proc_status");
        printf("wps_success_mode: %s, wps_status: %s\n",
               acosNvramConfig_get("wps_success_mode"),
               wsc_status);
        if(acosNvramConfig_match("wps_success_mode", "3"))
        {
            acosNvramConfig_set("wps_success_mode_2", "3"); /* for GUI. */
            acosNvramConfig_set("wps_success_mode", "0");

            
            acosNvramConfig_set("lan_wps_oob", "disabled");
            acosNvramConfig_set("wl_wps_config_state", "1");
            acosNvramConfig_set("wl0_wps_config_state", "1");
            

            printf("\n===============================\n");
            printf("For sync setting of AP of DUT and Cli of DUT., Copy wl0_XXX to wl0.1_XXX\n");
            printf("===============================\n");
            {/* SSID of AP of RPT */
                char ssid_ext[256];
                sprintf(ssid_ext, "%s_EXT", acosNvramConfig_get("wl0_ssid"));
                acosNvramConfig_set("wl0.1_ssid", ssid_ext);
            }
 #if 0
            /* Wifi Mode */
            acosNvramConfig_set("wla_mode", "300Mbps");
            acosNvramConfig_set("wl0.1_nbw_cap", acosNvramConfig_get("wl0_nbw_cap"));
            acosNvramConfig_set("wl0.1_nctrlsb", acosNvramConfig_get("wl0_nctrlsb"));
 #endif
            /* WEP */
            acosNvramConfig_set("wl0.1_wep",  acosNvramConfig_get("wl0_wep"));
            acosNvramConfig_set("wl0.1_key", acosNvramConfig_get("wl0_key"));
            acosNvramConfig_set("wl0.1_key1", acosNvramConfig_get("wl0_key1"));
            acosNvramConfig_set("wl0.1_key2", acosNvramConfig_get("wl0_key2"));
            acosNvramConfig_set("wl0.1_key3", acosNvramConfig_get("wl0_key3"));
            acosNvramConfig_set("wl0.1_key4", acosNvramConfig_get("wl0_key4"));
            acosNvramConfig_set("wl0.1_auth", "2");
            /* WPS */
            acosNvramConfig_set("wl0.1_akm", acosNvramConfig_get("wl0_akm"));
            acosNvramConfig_set("wl0.1_wpa_psk", acosNvramConfig_get("wl0_wpa_psk"));
            acosNvramConfig_set("wl0.1_crypto", acosNvramConfig_get("wl0_crypto"));
            nvram_commit();
        }/* end of wps_success_mode == 3 */

#ifdef WPS_DEBUG
    printf("\n");
    printf("============================================\n");
    printf("%s(), involed by rc restart.\n", __FUNCTION__);
    printf(" 1.wl0_akm:%s.\n",          nvram_get("wl0_akm"));
    printf("   wl0_ssid:%s.\n",         nvram_get("wl0_ssid"));
    printf("   wl0_wpa_psk:%s.\n",      nvram_get("wl0_wpa_psk"));
    printf("   wl0_crypto:%s.\n",       nvram_get("wl0_crypto"));
    printf("   wl0_wep:%s.\n",          nvram_get("wl0_wep"));
    printf("   wl0_auth:%s.\n",         nvram_get("wl0_auth"));
    printf("    \n");
    printf(" 2.wl0.1_akm:%s.\n",        nvram_get("wl0.1_akm"));
    printf("   wl0.1_ssid:%s.\n",       nvram_get("wl0.1_ssid"));
    printf("   wl0.1_wpa_psk:%s.\n",    nvram_get("wl0.1_wpa_psk"));
    printf("   wl0.1_crypto:%s.\n",     nvram_get("wl0.1_crypto"));
    printf("   wl0.1_wep:%s.\n",        nvram_get("wl0.1_wep"));
    printf("   wl0.1_auth:%s.\n",       nvram_get("wl0.1_auth"));
    printf("    \n");
    printf(" 3.wla_ssid_2:%s\n",    nvram_get("wla_ssid_2"));
    printf("   wla_secu_type_2:%s\n",   nvram_get("wla_secu_type_2"));
    printf(" 4.wps_randomssid:%s\n",    nvram_get("wps_randomssid"));
    printf("   wps_randomkey:%s\n",     nvram_get("wps_randomkey"));
    printf("\n\ncopy wl0_XXX to wla_XXX.\n");
    printf("copy wl0.1_XXX to wla_XXX_2.\n");
    printf("============================================\n");
#endif

#ifdef WIFI_DEBUG_MODE
    nvram_set("wl1.1_nas_dbg", "1");
    nvram_set("wl_nas_dbg", "1");
    nvram_set("wl0_nas_dbg", "1");
    nvram_set("wl1_nas_deb", "1");
    nvram_set("wl0.1_nas_dbg", "1");
    nvram_set("eapd_dbg", "3");
    nvram_set("nas_dbg", "1");
#endif

#ifdef WLAN_REPEATER
        nvram_set("enable_ether_counter_for_dhcpd", "0");
#endif

    nvram_set("wla_ssid_2", nvram_safe_get("wl0.1_ssid"));
    nvram_set("wla_temp_ssid_2", nvram_safe_get("wl0.1_ssid"));

    if ( strncmp(nvram_safe_get("wl0.1_akm"), "psk psk2", 7) == 0 )
    {/* WPA+WPA2 */
        nvram_set("wla_secu_type_2", "WPA-AUTO-PSK");
        nvram_set("wla_temp_secu_type_2", "WPA-AUTO-PSK");
        nvram_set("wla_passphrase_2", nvram_safe_get("wl0.1_wpa_psk"));
        if (!nvram_match("wps_randomssid", "") ||
            !nvram_match("wps_randomkey", ""))
        {
            //nvram_set("wla_secu_type_2", "WPA-AUTO-PSK");
            //nvram_set("wla_temp_secu_type_2","WPA-AUTO-PSK");
            //nvram_set("wl0.1_akm", "psk psk2 ");
            //nvram_set("wl0.1_crypto", "tkip+aes");
            nvram_set("wps_mixedmode", "2");
            nvram_set("wps_randomssid", "");
            nvram_set("wps_randomkey", "");
        }
    }else if(strncmp(nvram_safe_get("wl0.1_akm"), "psk2", 4) == 0 )
    {/* WPA2+AES */
        nvram_set("wla_secu_type_2", "WPA2-PSK");
        nvram_set("wla_temp_secu_type_2", "WPA2-PSK");
        nvram_set("wla_passphrase_2", nvram_safe_get("wl0.1_wpa_psk"));
        if (!nvram_match("wps_randomssid", "") ||
            !nvram_match("wps_randomkey", ""))
        {
            //nvram_set("wla_secu_type_2", "WPA-AUTO-PSK");
            //nvram_set("wla_temp_secu_type_2","WPA-AUTO-PSK");
            //nvram_set("wl0.1_akm", "psk psk2 ");
            //nvram_set("wl0.1_crypto", "tkip+aes");
            nvram_set("wps_mixedmode", "2"); /* If in mixed mode, defaultly
                                                using WPA2/AES */
            nvram_set("wps_randomssid", "");
            nvram_set("wps_randomkey", "");
        }

    }else if ( strncmp(nvram_safe_get("wl0.1_akm"), "psk", 3) == 0 )
    {/* WPA+TKIP */
        nvram_set("wla_secu_type_2", "WPA-PSK");
        nvram_set("wla_temp_secu_type_2", "WPA-PSK");
        nvram_set("wla_passphrase_2", nvram_safe_get("wl0.1_wpa_psk"));
        if (!nvram_match("wps_randomssid", "") ||
            !nvram_match("wps_randomkey", ""))
        {
            //nvram_set("wla_secu_type_2", "WPA-AUTO-PSK");
            //nvram_set("wla_temp_secu_type_2","WPA-AUTO-PSK");
            //nvram_set("wl0.1_akm", "psk psk2 ");
            //nvram_set("wl0.1_crypto", "tkip+aes");
            nvram_set("wps_mixedmode", "2");
            nvram_set("wps_randomssid", "");
            nvram_set("wps_randomkey", "");
        }
    }else if ( strncmp(nvram_safe_get("wl0.1_wep"), "enabled", 7) == 0 )
    {
        int key_len=0;
        int i;

        /* Some root AP give us key0, so we copy it to key1 properly */
        int wep_key_index, orig_wep_key_index;
        if (is_wps_sta_mode()) 
        {
            wep_key_index = atoi(nvram_safe_get("wl0.1_key"));
            orig_wep_key_index = wep_key_index;

            if (wep_key_index == 0)
            {
                char key1_nvram[128];
                /* Set the key index to 1 and copy key0 to key1 */
                printf("*** wep_key_index is 0, set to key1\n");
                nvram_set("wl0.1_key", "1");
                sprintf(key1_nvram, "%s", "wl0.1_key1");
                nvram_set(key1_nvram,
                          nvram_safe_get("wl0_key0"));
                wep_key_index = 1;
            }
        }



        nvram_set("wla_auth_type_2", "automatic");


        nvram_set("wla_secu_type_2", "WEP");
        nvram_set("wla_temp_secu_type_2", "WEP");
        //nvram_set("wla_defaKey_2", "0");
        //nvram_set("wla_temp_defaKey_2", "0");
        {
            int key;
            char value[8];
            char main_key[32];
            
            /* case 1: for sta mode wps 
                    * case 2: ap mode or normal setting 
                    * */
            if (is_wps_sta_mode()) 
            {
                /* The default key not always 0 (key1), set index properly */
                sprintf(value, "%d", wep_key_index-1);
                nvram_set("wla_defaKey_2", value);
                nvram_set("wla_temp_defaKey_2", value);
            }
            else
            {
                printf("wl0.1_key:%s\n", nvram_safe_get("wl0.1_key"));
                key = atoi(nvram_safe_get("wl0.1_key"));
                key -= 1;
                sprintf(value, "%d", key);
                //printf("value:%s\n", value);
                nvram_set("wla_defaKey_2", value);
                nvram_set("wla_temp_defaKey_2", value);
            }

            if(key != -1)
            {
                if (is_wps_sta_mode()) 
                {
                    /* It is not always WEP key 0. 
                                  * Should check NVRAM wlx_key for the correct index */
                    sprintf(main_key, "wl0.1_key%d", wep_key_index); 
                }
                else
                {
                    sprintf(main_key, "wl0.1_key%d", key+1);
                }
                
                key_len = strlen(nvram_safe_get(main_key));
                if (key_len==5 || key_len==10)
                    nvram_set("wla_wep_length_2", "1");
                else
                    nvram_set("wla_wep_length_2", "2");                
            }
        }

        if ( is_wps_sta_mode() )
        {
            /* Should read key1, after above "copy key0->key1" modification */
            /* nvram_set(get_output_nvram_name("_key1"), nvram_safe_get(get_input_nvram_name("_key0"))); */
            nvram_set("wla_key1_2", nvram_safe_get("wl0.1_key1"));

            /* If WEP key index is 0, we should additional copy
             * the wlx_key0 to * wlx.1_key1
             */
            if (orig_wep_key_index == 0 )
            {
                nvram_set("wl0.1_key1", nvram_safe_get("wl0.1_key1"));
            }

            /* Clear the passphrase, so that GUI won't show the WEP passphrase */
            nvram_set("wla_passphrase_2", "");
        }
        else
            nvram_set("wla_key1_2", nvram_safe_get("wl0.1_key1"));
        nvram_set("wla_key2_2", nvram_safe_get("wl0.1_key2"));
        nvram_set("wla_key3_2", nvram_safe_get("wl0.1_key3"));
        nvram_set("wla_key4_2", nvram_safe_get("wl0.1_key4"));

        if ( is_wps_sta_mode() )
            nvram_set("wla_temp_key1_2", nvram_safe_get("wl0.1_key0"));
        else
            nvram_set("wla_temp_key1_2", nvram_safe_get("wl0.1_key1"));
        nvram_set("wla_temp_key2_2", nvram_safe_get("wl0.1_key2"));
        nvram_set("wla_temp_key3_2", nvram_safe_get("wl0.1_key3"));
        nvram_set("wla_temp_key4_2", nvram_safe_get("wl0.1_key4"));
#ifdef DEBUG
        printf("wla_wep_length_2: %d\n", strlen(nvram_safe_get("wl0.1_key1")));
#endif
        for(i=1; i<5; i++){
            char key_no[64];
            char gui_key_no[64];
            char gui_tmp_key_no[64];
            memset(key_no, 0, sizeof(key_no));
            memset(gui_key_no, 0, sizeof(gui_key_no));
            memset(gui_tmp_key_no, 0, sizeof(gui_tmp_key_no));
            sprintf(key_no, "wl0.1_key%d", i);
            sprintf(gui_key_no, "wla_key%d_2", i);
            sprintf(gui_tmp_key_no, "wla_temp_key%d_2", i);
            //sprintf(key_len, "wla_wep_length_2");
            //key_len = strlen(nvram_safe_get("wl0.1_key1"));
       
            if (key_len==5 || key_len==13)
            {
                char HexKeyArray[32];
                char key[32], tmp[32];
                int i;

                //strcpy(key, nvram_safe_get("wl0.1_key1"));
                strcpy(key, nvram_safe_get(key_no));
                memset(HexKeyArray, 0, sizeof(HexKeyArray));
                for (i=0; i<key_len; i++)
                {
                    sprintf(tmp, "%02X", key[i]);
                    strcat(HexKeyArray, tmp);
                }
                printf("ASCII WEP key (%s) convert -> HEX WEP key (%s)\n", key, HexKeyArray);

                //nvram_set("wla_key1_2", HexKeyArray);
                //nvram_set("wla_temp_key1_2", HexKeyArray);
                nvram_set(gui_key_no, HexKeyArray);
                nvram_set(gui_tmp_key_no, HexKeyArray);
            }
        }
    }else
    {
        nvram_set("wla_secu_type_2", "None");
        nvram_set("wla_temp_secu_type_2", "None");
        nvram_set("wla_passphrase_2", "");
    }
#endif

    
    nvram_set("wla_ssid", nvram_safe_get("wl0_ssid"));
    nvram_set("wla_temp_ssid", nvram_safe_get("wl0_ssid"));

    if ( strncmp(nvram_safe_get("wl0_akm"), "psk psk2", 7) == 0 )
    {
        nvram_set("wla_secu_type", "WPA-AUTO-PSK");
        nvram_set("wla_temp_secu_type", "WPA-AUTO-PSK");
        nvram_set("wla_passphrase", nvram_safe_get("wl0_wpa_psk"));
#ifndef WLAN_REPEATER
        /* If router changes from 'unconfigured' to 'configured' state by
         * adding a WPS client, the wsc_randomssid and wsc_randomkey will
         * be set. In this case, router should use mixedmode security.
         */

        if (!nvram_match("wps_randomssid", "") ||
            !nvram_match("wps_randomkey", ""))
        {
            nvram_set("wla_secu_type", "WPA-AUTO-PSK");
            nvram_set("wla_temp_secu_type", "WPA-AUTO-PSK");

            nvram_set("wl0_akm", "psk psk2 ");
            nvram_set("wl0_crypto", "tkip+aes");

            nvram_set("wps_mixedmode", "2");
            nvram_set("wps_randomssid", "");
            nvram_set("wps_randomkey", "");

            /* Since we changed to mixed mode, 
             * so we need to disable WDS if it is already enabled
             */
            if (nvram_match("wla_wds_enable", "1"))
            {
                nvram_set("wla_wds_enable",  "0");
                nvram_set("wl0_wds", "");
                nvram_set("wl0_mode", "ap");
            }
        }
        else
        {
            /* Disable WDS if it is already enabled */
            if (nvram_match("wla_wds_enable", "1"))
            {
                nvram_set("wla_wds_enable",  "0");
                nvram_set("wl0_wds", "");
                nvram_set("wl0_mode", "ap");
            }
        }
#endif        
    }
    else if ( strncmp(nvram_safe_get("wl0_akm"), "psk2", 4) == 0 )
    {
        nvram_set("wla_secu_type", "WPA2-PSK");
        nvram_set("wla_temp_secu_type", "WPA2-PSK");
        nvram_set("wla_passphrase", nvram_safe_get("wl0_wpa_psk"));

#ifndef WLAN_REPEATER
        /* If router changes from 'unconfigured' to 'configured' state by
         * adding a WPS client, the wsc_randomssid and wsc_randomkey will
         * be set. In this case, router should use mixedmode security.
         */

        if (!nvram_match("wps_randomssid", "") ||
            !nvram_match("wps_randomkey", ""))
        {
            nvram_set("wla_secu_type", "WPA-AUTO-PSK");
            nvram_set("wla_temp_secu_type", "WPA-AUTO-PSK");

            nvram_set("wl0_akm", "psk psk2 ");
            nvram_set("wl0_crypto", "tkip+aes");

            nvram_set("wps_mixedmode", "2");
            nvram_set("wps_randomssid", "");
            nvram_set("wps_randomkey", "");

            /* Since we changed to mixed mode, 
             * so we need to disable WDS if it is already enabled
             */
            if (nvram_match("wla_wds_enable", "1"))
            {
                nvram_set("wla_wds_enable",  "0");
                nvram_set("wl0_wds", "");
                nvram_set("wl0_mode", "ap");
            }
        }
        else
        {
            /* Disable WDS if it is already enabled */
            if (nvram_match("wla_wds_enable", "1"))
            {
                nvram_set("wla_wds_enable",  "0");
                nvram_set("wl0_wds", "");
                nvram_set("wl0_mode", "ap");
            }
        }
#endif        
    }
    else if ( strncmp(nvram_safe_get("wl0_akm"), "psk", 3) == 0 )
    {
        nvram_set("wla_secu_type", "WPA-PSK");
        nvram_set("wla_temp_secu_type", "WPA-PSK");
        nvram_set("wla_passphrase", nvram_safe_get("wl0_wpa_psk"));
#ifndef WLAN_REPEATER
        /* If router changes from 'unconfigured' to 'configured' state by
         * adding a WPS client, the wsc_randomssid and wsc_randomkey will
         * be set. In this case, router should use mixedmode security.
         */

        if (!nvram_match("wps_randomssid", "") ||
            !nvram_match("wps_randomkey", ""))
        {
            nvram_set("wla_secu_type", "WPA-AUTO-PSK");
            nvram_set("wla_temp_secu_type", "WPA-AUTO-PSK");

            nvram_set("wl0_akm", "psk psk2 ");
            nvram_set("wl0_crypto", "tkip+aes");

            nvram_set("wps_mixedmode", "1");
            nvram_set("wps_randomssid", "");
            nvram_set("wps_randomkey", "");

            /* Since we changed to mixed mode, 
             * so we need to disable WDS if it is already enabled
             */
            if (nvram_match("wla_wds_enable", "1"))
            {
                nvram_set("wla_wds_enable",  "0");
                nvram_set("wl0_wds", "");
                nvram_set("wl0_mode", "ap");
            }
        }
#endif        
    }
    else if ( strncmp(nvram_safe_get("wl0_wep"), "enabled", 7) == 0 )
    {
        int key_len=0;

        /* Some root AP give us key0, so we copy it to key1 properly */
        int wep_key_index, orig_wep_key_index;
        if (is_wps_sta_mode()) 
        {
            wep_key_index = atoi(nvram_safe_get("wl0_key"));
            orig_wep_key_index = wep_key_index;

            if (wep_key_index == 0)
            {
                char key1_nvram[128];
                /* Set the key index to 1 and copy key0 to key1 */
                printf("*** wep_key_index is 0, set to key1\n");
                nvram_set("wl0_key", "1");
                sprintf(key1_nvram, "%s", "wl0_key1");
                nvram_set(key1_nvram,
                          nvram_safe_get("wl0_key0"));
                wep_key_index = 1;
            }
        }
        if ( strncmp(nvram_safe_get("wl0_auth"), "1", 1) == 0 ) /*shared mode*/
        {
            nvram_set("wla_auth_type", "sharedkey");
            nvram_set("wla_temp_auth_type", "sharedkey");
        }
        else
        {
            nvram_set("wla_auth_type", "opensystem");
            nvram_set("wla_temp_auth_type", "opensystem");
        }
        nvram_set("wla_secu_type", "WEP");
        nvram_set("wla_temp_secu_type", "WEP");
#ifdef WLAN_REPEATER   
        {
            int key;
            char value[8];
            char main_key[32];
            int key_length;
            /* case 1: for sta mode wps 
             * case 2: ap mode or normal setting 
             * */
            if ( is_wps_sta_mode() ) {
                /* The default key not always 0 (key1), set index properly */

                char def_key[4];
                sprintf(def_key, "%d", wep_key_index-1);
                nvram_set("wla_defaKey", def_key);
                nvram_set("wla_temp_defaKey", def_key);
            } else {   
	            printf("wl0_key: %s\n", nvram_safe_get("wl0_key"));
	            key = atoi(nvram_safe_get("wl0_key"));
	            key -= 1;
	            sprintf(value, "%d", key);
	            //printf("value:%s\n", value);
	            nvram_set("wla_defaKey", value);
	            nvram_set("wla_temp_defaKey", value);
			    /* Calculate the length of key using the main key. */
            }
            if(key != -1 )
            {
                if ( is_wps_sta_mode() ) 
                {
                    /* It is not always WEP key 0. 
                     * Should check NVRAM wlx_key for the correct index */
                    /*sprintf(main_key, "%s%d", get_input_nvram_name("_key"), 0);*/
                    sprintf(main_key, "wl0_key%d", wep_key_index); 
                    //printf("main_key='%s'\n", main_key);
                }
                else  
                    sprintf(main_key, "wl0_key%d", key+1);
                key_length = strlen(nvram_safe_get(main_key));
                printf("key_length: %d\n", key_length);
                if (key_len==5 || key_len==10)
                    nvram_set("wla_wep_length", "1");
                else
                    nvram_set("wla_wep_length", "2");
            }
        }
#else   
        nvram_set("wla_defaKey", "0");
        nvram_set("wla_temp_defaKey", "0");
#endif

        if ( is_wps_sta_mode() )
        {
            /* Should read key1, after above "copy key0->key1" modification */
            /* nvram_set(get_output_nvram_name("_key1"), nvram_safe_get(get_input_nvram_name("_key0"))); */
            nvram_set("wla_key1", nvram_safe_get("wl0_key1"));


            /* Clear the passphrase, so that GUI won't show the WEP passphrase */
            nvram_set("wla_passphrase", "");
        }
        else
            nvram_set("wla_key1", nvram_safe_get("wl0_key1"));
#ifdef WLAN_REPEATER
        nvram_set("wla_key2", nvram_safe_get("wl0_key2"));
        nvram_set("wla_key3", nvram_safe_get("wl0_key3"));
        nvram_set("wla_key4", nvram_safe_get("wl0_key4"));
#endif
        if ( is_wps_sta_mode() )
            nvram_set("wla_temp_key1", nvram_safe_get("wl0_key0"));
        else
            nvram_set("wla_temp_key1", nvram_safe_get("wl0_key1"));
#ifdef WLAN_REPEATER
        nvram_set("wla_temp_key2", nvram_safe_get("wl0_key2"));
        nvram_set("wla_temp_key3", nvram_safe_get("wl0_key3"));
        nvram_set("wla_temp_key4", nvram_safe_get("wl0_key4"));
#endif
        printf("wla_wep_length: %d\n", strlen(nvram_safe_get("wl0_key1")));

#ifdef WLAN_REEATER   
        for(i=1; i<5; i++){
            char key_no[64];
            char gui_key_no[64];
            char gui_tmp_key_no[64];
            memset(key_no, 0, sizeof(key_no));
            memset(gui_key_no, 0, sizeof(gui_key_no));
            memset(gui_tmp_key_no, 0, sizeof(gui_tmp_key_no));
            sprintf(key_no, "wl0_key%d", i);
            sprintf(gui_key_no, "wla_key%d", i);
            sprintf(gui_tmp_key_no, "wla_temp_key%d", i);
            //sprintf(key_len, "wla_wep_length_2");
            //key_len = strlen(nvram_safe_get("wl0.1_key1"));
            key_len = strlen(nvram_safe_get(key_no));

            if (key_len==5 || key_len==13)
            {
                char HexKeyArray[32];
                char key[32], tmp[32];
                int i;

                //strcpy(key, nvram_safe_get("wl0.1_key1"));
                strcpy(key, nvram_safe_get(key_no));
                memset(HexKeyArray, 0, sizeof(HexKeyArray));
                for (i=0; i<key_len; i++)
                {
                    sprintf(tmp, "%02X", key[i]);
                    strcat(HexKeyArray, tmp);
                }
                printf("ASCII WEP key (%s) convert -> HEX WEP key (%s)\n", key, HexKeyArray);

                //nvram_set("wla_key1_2", HexKeyArray);
                //nvram_set("wla_temp_key1_2", HexKeyArray);
                nvram_set(gui_key_no, HexKeyArray);
                nvram_set(gui_tmp_key_no, HexKeyArray);
            }
        }
#else	
        key_len = strlen(nvram_safe_get("wl0_key1"));
        
        if (key_len==5 || key_len==10)
        {
            nvram_set("wla_wep_length", "1");
        }
        else
        {
            nvram_set("wla_wep_length", "2");
        }
        if (key_len==5 || key_len==13)
        {
            char HexKeyArray[32];
            char key[32], tmp[32];
            int i;
            
            strcpy(key, nvram_safe_get("wl0_key1"));
            memset(HexKeyArray, 0, sizeof(HexKeyArray));
            for (i=0; i<key_len; i++)
            {
                sprintf(tmp, "%02X", key[i]);
                strcat(HexKeyArray, tmp);
            }
            printf("ASCII WEP key (%s) convert -> HEX WEP key (%s)\n", key, HexKeyArray);
            
            nvram_set("wla_key1", HexKeyArray);
            nvram_set("wla_temp_key1", HexKeyArray);
        }
#endif
    }
    else
    {
        nvram_set("wla_secu_type", "None");
        nvram_set("wla_temp_secu_type", "None");
        nvram_set("wla_passphrase", "");
    }

    nvram_set("allow_registrar_config", "0");  


    /* 'wl_unit' is changed to "0.-1" after Vista configure router (using Borg DTM1.3 patch).
     * This will make WPS fail to work on the correct interface.
     * Set it back to "0" if it is not.
     */
    if (!nvram_match("wl_unit", "0"))
        nvram_set("wl_unit", "0");
}

static void save_wlan_time(void)
{
    struct sysinfo info;
    char command[128];
    sysinfo(&info);
    sprintf(command, "echo %lu > /tmp/wlan_time", info.uptime);
    system(command);
    return;
}

static void qos_config_switch(void)
{
    /* BCM5325E & BCM5325F switch request to change these vars
     * to output ethernet port tag/id in packets.
     */
    struct nvram_tuple vlan_bcm4716[] = {
        { "wan_ifname", "vlan1", 0 },
        { "wan_ifnames", "vlan1 ", 0 },
        { "wandevs", "vlan1", 0 },
        { "vlan1ports", "4 5", 0 },
        { 0, 0, 0 }
    };

    //struct nvram_tuple *u = generic_bcm4716;
    struct nvram_tuple *u = vlan_bcm4716;
    int commit = 0;

    if (nvram_match("qos_enable", "1") && 
        !nvram_match("wla_repeater", "1") && 
        !nvram_match("qos_port", "")) {
        u = vlan_bcm4716;
    }

    for ( ; u && u->name; u++) {
        if (strcmp(nvram_get(u->name), u->value)) {
            commit = 1;
            nvram_set(u->name, u->value);
        }
    }

    if (commit) {
        printf("Commit new QoS switch config...\n");
        nvram_commit();
        commit = 0;
    }
}

static int send_wps_led_cmd(int cmd, int arg)
{
    int ret_val=0;
    int fd;

    fd = open(DEV_WPS_LED, O_RDWR);
    if (fd < 0) 
        return -1;
#ifdef WLAN_REPEATER 
    if (nvram_match("wla_secu_type_2", "None"))
#else
    if (nvram_match("wla_secu_type", "None"))
#endif
        arg = 0;
    else
        arg = 1;

    ret_val = ioctl(fd, cmd, arg);
    close(fd);

    return ret_val;
}

static int
build_ifnames(char *type, char *names, int *size)
{
	char name[32], *next;
	int len = 0;
	int s;

	/* open a raw scoket for ioctl */
	if ((s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
		return -1;

	/*
	 * go thru all device names (wl<N> il<N> et<N> vlan<N>) and interfaces to 
	 * build an interface name list in which each i/f name coresponds to a device
	 * name in device name list. Interface/device name matching rule is device
	 * type dependant:
	 *
	 *	wl:	by unit # provided by the driver, for example, if eth1 is wireless
	 *		i/f and its unit # is 0, then it will be in the i/f name list if
	 *		wl0 is in the device name list.
	 *	il/et:	by mac address, for example, if et0's mac address is identical to
	 *		that of eth2's, then eth2 will be in the i/f name list if et0 is 
	 *		in the device name list.
	 *	vlan:	by name, for example, vlan0 will be in the i/f name list if vlan0
	 *		is in the device name list.
	 */
	foreach(name, type, next) {
		struct ifreq ifr;
		int i, unit;
		char var[32], *mac;
		unsigned char ea[ETHER_ADDR_LEN];

		/* vlan: add it to interface name list */
		if (!strncmp(name, "vlan", 4)) {
			/* append interface name to list */
			len += snprintf(&names[len], *size - len, "%s ", name);
			continue;
		}

		/* others: proceed only when rules are met */
		for (i = 1; i <= DEV_NUMIFS; i ++) {
			/* ignore i/f that is not ethernet */
			ifr.ifr_ifindex = i;
			if (ioctl(s, SIOCGIFNAME, &ifr))
				continue;
			if (ioctl(s, SIOCGIFHWADDR, &ifr))
				continue;
			if (ifr.ifr_hwaddr.sa_family != ARPHRD_ETHER)
				continue;
			if (!strncmp(ifr.ifr_name, "vlan", 4))
				continue;

			/* wl: use unit # to identify wl */
			if (!strncmp(name, "wl", 2)) {
				if (wl_probe(ifr.ifr_name) ||
				    wl_ioctl(ifr.ifr_name, WLC_GET_INSTANCE, &unit, sizeof(unit)) ||
				    unit != atoi(&name[2]))
					continue;
			}
			/* et/il: use mac addr to identify et/il */
			else if (!strncmp(name, "et", 2) || !strncmp(name, "il", 2)) {
				snprintf(var, sizeof(var), "%smacaddr", name);
				if (!(mac = nvram_get(var)) || !ether_atoe(mac, ea) ||
				    bcmp(ea, ifr.ifr_hwaddr.sa_data, ETHER_ADDR_LEN))
					continue;
			}
			/* mac address: compare value */
			else if (ether_atoe(name, ea) &&
				!bcmp(ea, ifr.ifr_hwaddr.sa_data, ETHER_ADDR_LEN))
				;
			/* others: ignore */
			else
				continue;

			/* append interface name to list */
			len += snprintf(&names[len], *size - len, "%s ", ifr.ifr_name);
		}
	}

	close(s);

	*size = len;
	return 0;
}

static void
wps_restore_defaults(void)
{
	/* cleanly up nvram for WPS */
	nvram_unset("wps_seed");
	nvram_unset("wps_config_state");
	nvram_unset("wps_addER");
	nvram_unset("wps_device_pin");
	nvram_unset("wps_pbc_force");
	nvram_unset("wps_config_command");
	nvram_unset("wps_proc_status");
	nvram_unset("wps_status");
	nvram_unset("wps_method");
	nvram_unset("wps_proc_mac");
	nvram_unset("wps_sta_pin");
	nvram_unset("wps_currentband");
	nvram_unset("wps_restart");
	nvram_unset("wps_event");
	nvram_unset("wps_restart");

	nvram_unset("wps_enr_mode");
	nvram_unset("wps_enr_ifname");
	nvram_unset("wps_enr_ssid");
	nvram_unset("wps_enr_bssid");
	nvram_unset("wps_enr_wsec");

	nvram_unset("wps_unit");
}

static void
ses_cleanup(void)
{
	/* well known event to cleanly initialize state machine */
	nvram_set("ses_event", "2");

	/* Delete lethal dynamically generated variables */
	nvram_unset("ses_bridge_disable");
}

static void
ses_restore_defaults(void)
{
	char tmp[100], prefix[] = "wlXXXXXXXXXX_ses_";
	int i, j;
	int len = 0;

	/* Delete dynamically generated variables */
	for (i = 0; i < MAX_NVPARSE; i++) {
		sprintf(prefix, "wl%d_ses_", i);
		nvram_unset(strcat_r(prefix, "ssid", tmp));
		nvram_unset(strcat_r(prefix, "closed", tmp));
		nvram_unset(strcat_r(prefix, "wpa_psk", tmp));
		nvram_unset(strcat_r(prefix, "auth", tmp));
		nvram_unset(strcat_r(prefix, "wep", tmp));
		nvram_unset(strcat_r(prefix, "auth_mode", tmp));
		nvram_unset(strcat_r(prefix, "crypto", tmp));
		nvram_unset(strcat_r(prefix, "akm", tmp));
		nvram_unset(strcat_r(prefix, "wds", tmp));
		nvram_unset(strcat_r(prefix, "wds_timeout", tmp));
	}

	/* Delete dynamically generated variables */
	for (i = 0; i < DEV_NUMIFS; i++) {
		sprintf(prefix, "wl%d_wds", i);
		len = strlen(prefix);
		for (j = 0; j < MAX_NVPARSE; j++) {
			sprintf(&prefix[len], "%d", j);
			nvram_unset(prefix);
		}
	}

	nvram_unset("ses_fsm_current_states");
	nvram_unset("ses_fsm_last_status");
}
static void
virtual_radio_restore_defaults(void)
{
	char tmp[100], prefix[] = "wlXXXXXXXXXX_mssid_";
	int i, j;

	nvram_unset("unbridged_ifnames");
	nvram_unset("ure_disable");

	/* Delete dynamically generated variables */
	for (i = 0; i < MAX_NVPARSE; i++) {
		sprintf(prefix, "wl%d_", i);
		nvram_unset(strcat_r(prefix, "vifs", tmp));
		nvram_unset(strcat_r(prefix, "ssid", tmp));
		nvram_unset(strcat_r(prefix, "guest", tmp));
		nvram_unset(strcat_r(prefix, "ure", tmp));
		nvram_unset(strcat_r(prefix, "ipconfig_index", tmp));
		nvram_unset(strcat_r(prefix, "nas_dbg", tmp));
		sprintf(prefix, "lan%d_", i);
		nvram_unset(strcat_r(prefix, "ifname", tmp));
		nvram_unset(strcat_r(prefix, "ifnames", tmp));
		nvram_unset(strcat_r(prefix, "gateway", tmp));
		nvram_unset(strcat_r(prefix, "proto", tmp));
		nvram_unset(strcat_r(prefix, "ipaddr", tmp));
		nvram_unset(strcat_r(prefix, "netmask", tmp));
		nvram_unset(strcat_r(prefix, "lease", tmp));
		nvram_unset(strcat_r(prefix, "stp", tmp));
		nvram_unset(strcat_r(prefix, "hwaddr", tmp));
		sprintf(prefix, "dhcp%d_", i);
		nvram_unset(strcat_r(prefix, "start", tmp));
		nvram_unset(strcat_r(prefix, "end", tmp));

		/* clear virtual versions */
		for (j = 0; j < 16; j++) {
			sprintf(prefix, "wl%d.%d_", i, j);
			nvram_unset(strcat_r(prefix, "ssid", tmp));
			nvram_unset(strcat_r(prefix, "ipconfig_index", tmp));
			nvram_unset(strcat_r(prefix, "guest", tmp));
			nvram_unset(strcat_r(prefix, "closed", tmp));
			nvram_unset(strcat_r(prefix, "wpa_psk", tmp));
			nvram_unset(strcat_r(prefix, "auth", tmp));
			nvram_unset(strcat_r(prefix, "wep", tmp));
			nvram_unset(strcat_r(prefix, "auth_mode", tmp));
			nvram_unset(strcat_r(prefix, "crypto", tmp));
			nvram_unset(strcat_r(prefix, "akm", tmp));
			nvram_unset(strcat_r(prefix, "hwaddr", tmp));
			nvram_unset(strcat_r(prefix, "bss_enabled", tmp));
			nvram_unset(strcat_r(prefix, "bss_maxassoc", tmp));
			nvram_unset(strcat_r(prefix, "wme_bss_disable", tmp));
			nvram_unset(strcat_r(prefix, "ifname", tmp));
			nvram_unset(strcat_r(prefix, "unit", tmp));
			nvram_unset(strcat_r(prefix, "ap_isolate", tmp));
			nvram_unset(strcat_r(prefix, "macmode", tmp));
			nvram_unset(strcat_r(prefix, "maclist", tmp));
			nvram_unset(strcat_r(prefix, "maxassoc", tmp));
			nvram_unset(strcat_r(prefix, "mode", tmp));
			nvram_unset(strcat_r(prefix, "radio", tmp));
			nvram_unset(strcat_r(prefix, "radius_ipaddr", tmp));
			nvram_unset(strcat_r(prefix, "radius_port", tmp));
			nvram_unset(strcat_r(prefix, "radius_key", tmp));
			nvram_unset(strcat_r(prefix, "key", tmp));
			nvram_unset(strcat_r(prefix, "key1", tmp));
			nvram_unset(strcat_r(prefix, "key2", tmp));
			nvram_unset(strcat_r(prefix, "key3", tmp));
			nvram_unset(strcat_r(prefix, "key4", tmp));
			nvram_unset(strcat_r(prefix, "wpa_gtk_rekey", tmp));
			nvram_unset(strcat_r(prefix, "nas_dbg", tmp));
		}
	}
}
static void
ses_cl_cleanup(void)
{
	/* well known event to cleanly initialize state machine */
	nvram_set("ses_cl_event", "0");
}

#ifdef __CONFIG_NAT__
static void
auto_bridge(void)
{

	struct nvram_tuple generic[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "eth0 eth2 eth3 eth4", 0 },
		{ "wan_ifname", "eth1", 0 },
		{ "wan_ifnames", "eth1", 0 },
		{ 0, 0, 0 }
	};
#ifdef __CONFIG_VLAN__
	struct nvram_tuple vlan[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "vlan0 eth1 eth2 eth3", 0 },
		{ "wan_ifname", "vlan1", 0 },
		{ "wan_ifnames", "vlan1", 0 },
		{ 0, 0, 0 }
	};
#endif	/* __CONFIG_VLAN__ */
	struct nvram_tuple dyna[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "", 0 },
		{ "wan_ifname", "", 0 },
		{ "wan_ifnames", "", 0 },
		{ 0, 0, 0 }
	};
	struct nvram_tuple generic_auto_bridge[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "eth0 eth1 eth2 eth3 eth4", 0 },
		{ "wan_ifname", "", 0 },
		{ "wan_ifnames", "", 0 },
		{ 0, 0, 0 }
	};
#ifdef __CONFIG_VLAN__
	struct nvram_tuple vlan_auto_bridge[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "vlan0 vlan1 eth1 eth2 eth3", 0 },
		{ "wan_ifname", "", 0 },
		{ "wan_ifnames", "", 0 },
		{ 0, 0, 0 }
	};
#endif	/* __CONFIG_VLAN__ */

	struct nvram_tuple dyna_auto_bridge[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "", 0 },
		{ "wan_ifname", "", 0 },
		{ "wan_ifnames", "", 0 },
		{ 0, 0, 0 }
	};

	struct nvram_tuple *linux_overrides;
	struct nvram_tuple *t, *u;
	int auto_bridge = 0, i;
#ifdef __CONFIG_VLAN__
	uint boardflags;
#endif	/* __CONFIG_VLAN_ */
	char *landevs, *wandevs;
	char lan_ifnames[128], wan_ifnames[128];
	char dyna_auto_ifnames[128];
	char wan_ifname[32], *next;
	int len;
	int ap = 0;

	printf(" INFO : enter function auto_bridge()\n");

	if (!strcmp(nvram_safe_get("auto_bridge_action"), "1")) {
		auto_bridge = 1;
		cprintf("INFO: Start auto bridge...\n");
	} else {
		nvram_set("router_disable_auto", "0");
		cprintf("INFO: Start non auto_bridge...\n");
	}

	/* Delete dynamically generated variables */
	if (auto_bridge) {
		char tmp[100], prefix[] = "wlXXXXXXXXXX_";
		for (i = 0; i < MAX_NVPARSE; i++) {

			del_filter_client(i);
			del_forward_port(i);
			del_autofw_port(i);


			snprintf(prefix, sizeof(prefix), "wan%d_", i);
			for (t = router_defaults; t->name; t ++) {
				if (!strncmp(t->name, "wan_", 4))
					nvram_unset(strcat_r(prefix, &t->name[4], tmp));
			}
		}
	}

	/* 
	 * Build bridged i/f name list and wan i/f name list from lan device name list
	 * and wan device name list. Both lan device list "landevs" and wan device list
	 * "wandevs" must exist in order to preceed.
	 */
	if ((landevs = nvram_get("landevs")) && (wandevs = nvram_get("wandevs"))) {
		/* build bridged i/f list based on nvram variable "landevs" */
		len = sizeof(lan_ifnames);
		if (!build_ifnames(landevs, lan_ifnames, &len) && len)
			dyna[1].value = lan_ifnames;
		else
			goto canned_config;
		/* build wan i/f list based on nvram variable "wandevs" */
		len = sizeof(wan_ifnames);
		if (!build_ifnames(wandevs, wan_ifnames, &len) && len) {
			dyna[3].value = wan_ifnames;
			foreach(wan_ifname, wan_ifnames, next) {
				dyna[2].value = wan_ifname;
				break;
			}
		}
		else
			ap = 1;

		if (auto_bridge)
		{
			printf("INFO: lan_ifnames=%s\n", lan_ifnames);
			printf("INFO: wan_ifnames=%s\n", wan_ifnames);
			sprintf(dyna_auto_ifnames, "%s %s", lan_ifnames, wan_ifnames);
			printf("INFO: dyna_auto_ifnames=%s\n", dyna_auto_ifnames);
			dyna_auto_bridge[1].value = dyna_auto_ifnames;
			linux_overrides = dyna_auto_bridge;
			printf("INFO: linux_overrides=dyna_auto_bridge \n");
		}
		else
		{
			linux_overrides = dyna;
			printf("INFO: linux_overrides=dyna \n");
		}

	}
	/* override lan i/f name list and wan i/f name list with default values */
	else {
canned_config:
#ifdef __CONFIG_VLAN__
		boardflags = strtoul(nvram_safe_get("boardflags"), NULL, 0);
		if (boardflags & BFL_ENETVLAN) {
			if (auto_bridge)
			{
				linux_overrides = vlan_auto_bridge;
				printf("INFO: linux_overrides=vlan_auto_bridge \n");
			}
			else
			{
				linux_overrides = vlan;
				printf("INFO: linux_overrides=vlan \n");
			}
		} else {
#endif	/* __CONFIG_VLAN__ */
			if (auto_bridge)
			{
				linux_overrides = generic_auto_bridge;
				printf("INFO: linux_overrides=generic_auto_bridge \n");
			}
			else
			{
				linux_overrides = generic;
				printf("INFO: linux_overrides=generic \n");
			}
#ifdef __CONFIG_VLAN__
		}
#endif	/* __CONFIG_VLAN__ */
	}

		for (u = linux_overrides; u && u->name; u++) {
			nvram_set(u->name, u->value);
			printf("INFO: action nvram_set %s, %s\n", u->name, u->value);
			}

	/* Force to AP */
	if (ap)
		nvram_set("router_disable", "1");

	if (auto_bridge) {
		printf("INFO: reset auto_bridge flag.\n");
		nvram_set("auto_bridge_action", "0");
	}

	nvram_commit();
	cprintf("auto_bridge done\n");
}

#endif	/* __CONFIG_NAT__ */


static void
upgrade_defaults(void)
{
	char temp[100];
	int i;
	bool bss_enabled = TRUE;
	char *val;

	/* Check whether upgrade is required or not
	 * If lan1_ifnames is not found in NVRAM , upgrade is required.
	 */
	if (!nvram_get("lan1_ifnames") && !RESTORE_DEFAULTS()) {
		cprintf("NVRAM upgrade required.  Starting.\n");

		if (nvram_match("ure_disable", "1")) {
			nvram_set("lan1_ifname", "br1");
			nvram_set("lan1_ifnames", "wl0.1 wl0.2 wl0.3 wl1.1 wl1.2 wl1.3");
		}
		else {
			nvram_set("lan1_ifname", "");
			nvram_set("lan1_ifnames", "");
			for (i = 0; i < 2; i++) {
				snprintf(temp, sizeof(temp), "wl%d_ure", i);
				if (nvram_match(temp, "1")) {
					snprintf(temp, sizeof(temp), "wl%d.1_bss_enabled", i);
					nvram_set(temp, "1");
				}
				else {
					bss_enabled = FALSE;
					snprintf(temp, sizeof(temp), "wl%d.1_bss_enabled", i);
					nvram_set(temp, "0");
				}
			}
		}
		if (nvram_get("lan1_ipaddr")) {
			nvram_set("lan1_gateway", nvram_get("lan1_ipaddr"));
		}

		for (i = 0; i < 2; i++) {
			snprintf(temp, sizeof(temp), "wl%d_bss_enabled", i);
			nvram_set(temp, "1");
			snprintf(temp, sizeof(temp), "wl%d.1_guest", i);
			if (nvram_match(temp, "1")) {
				nvram_unset(temp);
				if (bss_enabled) {
					snprintf(temp, sizeof(temp), "wl%d.1_bss_enabled", i);
					nvram_set(temp, "1");
				}
			}

			snprintf(temp, sizeof(temp), "wl%d.1_net_reauth", i);
			val = nvram_get(temp);
			if (!val || (*val == 0))
				nvram_set(temp, nvram_default_get(temp));

			snprintf(temp, sizeof(temp), "wl%d.1_wpa_gtk_rekey", i);
			val = nvram_get(temp);
			if (!val || (*val == 0))
				nvram_set(temp, nvram_default_get(temp));
		}

		nvram_commit();

		cprintf("NVRAM upgrade complete.\n");
	}
}

static void
restore_defaults(void)
{
#if 0 
	struct nvram_tuple generic[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "eth0 eth2 eth3 eth4", 0 },
		{ "wan_ifname", "eth1", 0 },
		{ "wan_ifnames", "eth1", 0 },
		{ "lan1_ifname", "br1", 0 },
		{ "lan1_ifnames", "wl0.1 wl0.2 wl0.3 wl1.1 wl1.2 wl1.3", 0 },
		{ 0, 0, 0 }
	};
#ifdef __CONFIG_VLAN__
	struct nvram_tuple vlan[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "vlan0 eth1 eth2 eth3", 0 },
		{ "wan_ifname", "vlan1", 0 },
		{ "wan_ifnames", "vlan1", 0 },
		{ "lan1_ifname", "br1", 0 },
		{ "lan1_ifnames", "wl0.1 wl0.2 wl0.3 wl1.1 wl1.2 wl1.3", 0 },
		{ 0, 0, 0 }
	};
#endif	/* __CONFIG_VLAN__ */
	struct nvram_tuple dyna[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "", 0 },
		{ "wan_ifname", "", 0 },
		{ "wan_ifnames", "", 0 },
		{ "lan1_ifname", "br1", 0 },
		{ "lan1_ifnames", "wl0.1 wl0.2 wl0.3 wl1.1 wl1.2 wl1.3", 0 },
		{ 0, 0, 0 }
	};
#endif /* 0 */ 

	/* struct nvram_tuple *linux_overrides; */
	struct nvram_tuple *t, *u;
	int restore_defaults, i;
#ifdef __CONFIG_VLAN__
	uint boardflags;
#endif	/* __CONFIG_VLAN_ */
    /*
	char *landevs, *wandevs;
	char lan_ifnames[128], wan_ifnames[128];
	char wan_ifname[32], *next;
	int len;
	int ap = 0;
    */

	/* Restore defaults if told to or OS has changed */
	restore_defaults = RESTORE_DEFAULTS();

	if (restore_defaults)
		cprintf("Restoring defaults...");

	/* Delete dynamically generated variables */
	if (restore_defaults) {
		char tmp[100], prefix[] = "wlXXXXXXXXXX_";
		for (i = 0; i < MAX_NVPARSE; i++) {
#ifdef __CONFIG_NAT__
			del_filter_client(i);
			del_forward_port(i);
			del_autofw_port(i);
#endif	/* __CONFIG_NAT__ */
			snprintf(prefix, sizeof(prefix), "wl%d_", i);
			for (t = router_defaults; t->name; t ++) {
				if (!strncmp(t->name, "wl_", 3))
					nvram_unset(strcat_r(prefix, &t->name[3], tmp));
			}
#ifdef __CONFIG_NAT__
			snprintf(prefix, sizeof(prefix), "wan%d_", i);
			for (t = router_defaults; t->name; t ++) {
				if (!strncmp(t->name, "wan_", 4))
					nvram_unset(strcat_r(prefix, &t->name[4], tmp));
			}
#endif	/* __CONFIG_NAT__ */
		}
		ses_restore_defaults();
		wps_restore_defaults();
#ifdef __CONFIG_WAPI_IAS__
		nvram_unset("as_mode");
#endif /* __CONFIG_WAPI_IAS__ */
		virtual_radio_restore_defaults();
	}

#if 0 
	/* 
	 * Build bridged i/f name list and wan i/f name list from lan device name list
	 * and wan device name list. Both lan device list "landevs" and wan device list
	 * "wandevs" must exist in order to preceed.
	 */
	if ((landevs = nvram_get("landevs")) && (wandevs = nvram_get("wandevs"))) {
		/* build bridged i/f list based on nvram variable "landevs" */
		len = sizeof(lan_ifnames);
		if (!build_ifnames(landevs, lan_ifnames, &len) && len)
			dyna[1].value = lan_ifnames;
		else
			goto canned_config;
		/* build wan i/f list based on nvram variable "wandevs" */
		len = sizeof(wan_ifnames);
		if (!build_ifnames(wandevs, wan_ifnames, &len) && len) {
			dyna[3].value = wan_ifnames;
			foreach(wan_ifname, wan_ifnames, next) {
				dyna[2].value = wan_ifname;
				break;
			}
		}
		else
			ap = 1;
		linux_overrides = dyna;
	}
	/* override lan i/f name list and wan i/f name list with default values */
	else {
canned_config:
#ifdef __CONFIG_VLAN__
		boardflags = strtoul(nvram_safe_get("boardflags"), NULL, 0);
		if (boardflags & BFL_ENETVLAN)
			linux_overrides = vlan;
		else
#endif	/* __CONFIG_VLAN__ */
			linux_overrides = generic;
	}

	/* Check if nvram version is set, but old */
	if (nvram_get("nvram_version")) {
		int old_ver, new_ver;

		old_ver = atoi(nvram_get("nvram_version"));
		new_ver = atoi(NVRAM_SOFTWARE_VERSION);
		if (old_ver < new_ver) {
			cprintf("NVRAM: Updating from %d to %d\n", old_ver, new_ver);
			nvram_set("nvram_version", NVRAM_SOFTWARE_VERSION);
		}
	}
#endif /* 0 */

	/* Restore defaults */
	for (t = router_defaults; t->name; t++) {
		if (restore_defaults || !nvram_get(t->name)) {
#if 0 
			for (u = linux_overrides; u && u->name; u++) {
				if (!strcmp(t->name, u->name)) {
					nvram_set(u->name, u->value);
					break;
				}
			}
			if (!u || !u->name)
#endif
				nvram_set(t->name, t->value);
		}
	}

	/* Force to AP */
#if 0 
	if (ap)
		nvram_set("router_disable", "1");
#endif

	/* Always set OS defaults */
	nvram_set("os_name", "linux");
	nvram_set("os_version", ROUTER_VERSION_STR);
	nvram_set("os_date", __DATE__);
	/* Always set WL driver version! */
	nvram_set("wl_version", EPI_VERSION_STR);

	nvram_set("is_modified", "0");
	nvram_set("ezc_version", EZC_VERSION_STR);

    /* not allow wl driver control these GPIOs */
    /*nvram_set("ledbh1", "11");
    nvram_set("ledbh2", "11");
    nvram_set("ledbh3", "11");*/
	/* Commit values */
	if (restore_defaults) {
#ifdef FW_VERSION_NA
		acosNvramConfig_setPAParam(0);
#else
		acosNvramConfig_setPAParam(1);
#endif
		
        eval("read_bd"); 
		nvram_commit();
        /* Create a new value to inform loaddefault in ap\acos\rc\main.c */
        nvram_set("load_defaults", "1");
        sync();         
		cprintf("done\n");
	}

}

#ifdef __CONFIG_NAT__
static void
set_wan0_vars(void)
{
	int unit;
	char tmp[100], prefix[] = "wanXXXXXXXXXX_";

	/* check if there are any connections configured */
	for (unit = 0; unit < MAX_NVPARSE; unit ++) {
		snprintf(prefix, sizeof(prefix), "wan%d_", unit);
		if (nvram_get(strcat_r(prefix, "unit", tmp)))
			break;
	}
	/* automatically configure wan0_ if no connections found */
	if (unit >= MAX_NVPARSE) {
		struct nvram_tuple *t;
		char *v;

		/* Write through to wan0_ variable set */
		snprintf(prefix, sizeof(prefix), "wan%d_", 0);
		for (t = router_defaults; t->name; t ++) {
			if (!strncmp(t->name, "wan_", 4)) {
				if (nvram_get(strcat_r(prefix, &t->name[4], tmp)))
					continue;
				v = nvram_get(t->name);
				nvram_set(tmp, v ? v : t->value);
			}
		}
		nvram_set(strcat_r(prefix, "unit", tmp), "0");
		nvram_set(strcat_r(prefix, "desc", tmp), "Default Connection");
		nvram_set(strcat_r(prefix, "primary", tmp), "1");
	}
}
#endif	/* __CONFIG_NAT__ */


static int noconsole = 0;

static void
sysinit(void)
{
	char buf[PATH_MAX];
	struct utsname name;
	struct stat tmp_stat;
	time_t tm = 0;
	char *loglevel;

	/* /proc */
	mount("proc", "/proc", "proc", MS_MGC_VAL, NULL);
#ifdef LINUX26
	mount("sysfs", "/sys", "sysfs", MS_MGC_VAL, NULL);
#endif /* LINUX26 */

	/* /tmp */
	mount("ramfs", "/tmp", "ramfs", MS_MGC_VAL, NULL);

	/* /var */
	mkdir("/tmp/var", 0777);
	mkdir("/var/lock", 0777);
	mkdir("/var/log", 0777);
	mkdir("/var/run", 0777);
	mkdir("/var/tmp", 0777);
	mkdir("/tmp/media", 0777);

#ifdef __CONFIG_SAMBA__
	/* Add Samba Stuff */
	mkdir("/tmp/samba", 0777);
	mkdir("/tmp/samba/lib", 0777);
	mkdir("/tmp/samba/private", 0777);
	mkdir("/tmp/samba/var", 0777);
	mkdir("/tmp/samba/var/locks", 0777);
#endif

#ifdef BCMQOS
	mkdir("/tmp/qos", 0777);
#endif
	/* Setup console */
	if (console_init())
		noconsole = 1;

#ifdef LINUX26
	mkdir("/dev/shm", 0777);
	eval("/sbin/hotplug2", "--coldplug");
#endif /* LINUX26 */

	if ((loglevel = nvram_get("console_loglevel")))
		klogctl(8, NULL, atoi(loglevel));
	else
		klogctl(8, NULL, 1);

	/* Modules */
	uname(&name);
	snprintf(buf, sizeof(buf), "/lib/modules/%s", name.release);
	if (stat("/proc/modules", &tmp_stat) == 0 &&
	    stat(buf, &tmp_stat) == 0) {
		char module[80], *modules, *next;
		/* Load ctf */
		if (!nvram_match("ctf_disable", "1"))
			eval("insmod", "ctf");
#if defined(__CONFIG_WAPI__) || defined(__CONFIG_WAPI_IAS__)
		wapi_mtd_restore();
#endif /* __CONFIG_WAPI__ || __CONFIG_WAPI_IAS__ */


/* #ifdef BCMVISTAROUTER */
#ifdef __CONFIG_IPV6__
		eval("insmod", "ipv6");
#endif /* __CONFIG_IPV6__ */
/* #endif */

#ifdef __CONFIG_EMF__
		/* Load the EMF & IGMP Snooper modules */
		load_emf();
#endif /*  __CONFIG_EMF__ */

		modules = nvram_get("kernel_mods") ? : "et wl"; 

		foreach(module, modules, next)
			eval("insmod", module);

#ifdef __CONFIG_USBAP__
		/* We have to load USB modules after loading PCI wl driver so
		 * USB driver can decide its instance number based on PCI wl
		 * instance numbers (in hotplug_usb())
		 */
		eval("insmod", "usbcore");
		eval("insmod", "usb-storage");
		{
			char	insmod_arg[128];
			int	i = 0, maxwl_eth = 0, maxunit = -1;
			char	ifname[16] = {0};
			int	unit = -1;
			char arg1[20] = {0};
			char arg2[20] = {0};
			char arg3[20] = {0};
			char arg4[20] = {0};
			char arg5[20] = {0};
			char arg6[20] = {0};
			char arg7[20] = {0};
			const int wl_wait = 3;	/* max wait time for wl_high to up */

			/* Save QTD cache params in nvram */
			sprintf(arg1, "log2_irq_thresh=%d", atoi(nvram_safe_get(
				"log2_irq_thresh")));
			sprintf(arg2, "qtdc_pid=%d", atoi(nvram_safe_get("qtdc_pid")));
			sprintf(arg3, "qtdc_vid=%d", atoi(nvram_safe_get("qtdc_vid")));
			sprintf(arg4, "qtdc0_ep=%d", atoi(nvram_safe_get("qtdc0_ep")));
			sprintf(arg5, "qtdc0_sz=%d", atoi(nvram_safe_get("qtdc0_sz")));
			sprintf(arg6, "qtdc1_ep=%d", atoi(nvram_safe_get("qtdc1_ep")));
			sprintf(arg7, "qtdc1_sz=%d", atoi(nvram_safe_get("qtdc1_sz")));

			eval("insmod", "ehci-hcd", arg1, arg2, arg3, arg4, arg5,
				arg6, arg7);

			/* Search for existing PCI wl devices and the max unit number used.
			 * Note that PCI driver has to be loaded before USB hotplug event.
			 * This is enforced in rc.c
			 */
			for (i = 1; i <= DEV_NUMIFS; i++) {
				sprintf(ifname, "eth%d", i);
				if (!wl_probe(ifname)) {
					if (!wl_ioctl(ifname, WLC_GET_INSTANCE, &unit,
						sizeof(unit))) {
						maxwl_eth = i;
						maxunit = (unit > maxunit) ? unit : maxunit;
					}
				}
			}

			/* Set instance base (starting unit number) for USB device */
			sprintf(insmod_arg, "instance_base=%d", maxunit + 1);
			eval("insmod", "wl_high", insmod_arg);

			/* Hold until the USB/HSIC interface is up (up to wl_wait sec) */
			sprintf(ifname, "eth%d", maxwl_eth + 1);
			i = wl_wait;
			while (wl_probe(ifname) && i--) {
				sleep(1);
			}
			if (!wl_ioctl(ifname, WLC_GET_INSTANCE, &unit, sizeof(unit)))
				cprintf("wl%d is up in %d sec\n", unit, wl_wait - i);
			else
				cprintf("wl%d not up in %d sec\n", unit, wl_wait);
		}
#ifdef LINUX26
		mount("usbdeffs", "/proc/bus/usb", "usbfs", MS_MGC_VAL, NULL);
#else
		mount("none", "/proc/bus/usb", "usbdevfs", MS_MGC_VAL, NULL);
#endif /* LINUX26 */
#endif /* __CONFIG_USBAP__ */

#ifdef __CONFIG_WCN__
		modules = "scsi_mod sd_mod usbcore usb-ohci usb-storage fat vfat msdos";
		foreach(module, modules, next)
			eval("insmod", module);
#endif

#ifdef __CONFIG_SOUND__
		modules = "soundcore snd snd-timer snd-page-alloc snd-pcm snd-pcm-oss "
		        "snd-soc-core i2c-core i2c-algo-bit i2c-gpio snd-soc-bcm947xx-i2s "
		        "snd-soc-bcm947xx-pcm snd-soc-wm8750 snd-soc-wm8955 snd-soc-bcm947xx";
		foreach(module, modules, next)
			eval("insmod", module);
		mknod("/dev/dsp", S_IRWXU|S_IFCHR, makedev(14, 3));
		mkdir("/dev/snd", 0777);
		mknod("/dev/snd/controlC0", S_IRWXU|S_IFCHR, makedev(116, 0));
		mknod("/dev/snd/pcmC0D0c", S_IRWXU|S_IFCHR, makedev(116, 24));
		mknod("/dev/snd/pcmC0D0p", S_IRWXU|S_IFCHR, makedev(116, 16));
		mknod("/dev/snd/timer", S_IRWXU|S_IFCHR, makedev(116, 33));
#endif
	}

	/* Set a sane date */
	stime(&tm);

	dprintf("done\n");
}

/* States */
enum {
	RESTART,
	STOP,
	START,
	TIMER,
	IDLE,
	WSC_RESTART,
	WLANRESTART, 
    WLAN_SLOW_START 
};
static int state = START;
static int signalled = -1;

/* Signal handling */
static void
rc_signal(int sig)
{
	if (state == IDLE) {	
		if (sig == SIGHUP) {
			dprintf("signalling RESTART\n");
			signalled = RESTART;
		}
		else if (sig == SIGUSR2) {
			dprintf("signalling START\n");
			signalled = START;
		}
		else if (sig == SIGINT) {
			dprintf("signalling STOP\n");
			signalled = STOP;
		}
		else if (sig == SIGALRM) {
			dprintf("signalling TIMER\n");
			signalled = TIMER;
		}
		else if (sig == SIGUSR1) {
			dprintf("signalling WSC RESTART\n");
			signalled = WSC_RESTART;
		}
		else if (sig == SIGQUIT) {
		    printf("signalling WLANRESTART\n");
		    signalled = WLANRESTART;
		}
#ifdef WLAN_REPEATER
        else if (sig == SIGILL){
            dprintf("signalling WLAN_SLOW_START");
            signalled = WLAN_SLOW_START;
        }
#endif
	}
}

/* Get the timezone from NVRAM and set the timezone in the kernel
 * and export the TZ variable 
 */
static void
set_timezone(void)
{
	time_t now;
	struct tm gm, local;
	struct timezone tz;
	struct timeval *tvp = NULL;

	/* Export TZ variable for the time libraries to 
	 * use.
	 */
	setenv("TZ", nvram_get("time_zone"), 1);

	/* Update kernel timezone */
	time(&now);
	gmtime_r(&now, &gm);
	localtime_r(&now, &local);
	tz.tz_minuteswest = (mktime(&gm) - mktime(&local)) / 60;
	settimeofday(tvp, &tz);

#if defined(__CONFIG_WAPI__) || defined(__CONFIG_WAPI_IAS__)
#ifndef	RC_BUILDTIME
#define	RC_BUILDTIME	1252636574
#endif
	{
		struct timeval tv = {RC_BUILDTIME, 0};

		time(&now);
		if (now < RC_BUILDTIME)
			settimeofday(&tv, &tz);
	}
#endif /* __CONFIG_WAPI__ || __CONFIG_WAPI_IAS__ */
}

/* Timer procedure.Gets time from the NTP servers once every timer interval 
 * Interval specified by the NVRAM variable timer_interval
 */
int
do_timer(void)
{
	int interval = atoi(nvram_safe_get("timer_interval"));

	dprintf("%d\n", interval);

	if (interval == 0)
		return 0;

	/* Report stats */
	if (nvram_invmatch("stats_server", "")) {
		char *stats_argv[] = { "stats", nvram_get("stats_server"), NULL };
		_eval(stats_argv, NULL, 5, NULL);
	}

	/* Sync time */
	start_ntpc();

	alarm(interval);

	return 0;
}

/* Main loop */
static void
main_loop(void)
{
#ifdef CAPI_AP
	static bool start_aput = TRUE;
#endif
	sigset_t sigset;
	pid_t shell_pid = 0;
#ifdef __CONFIG_VLAN__
	uint boardflags;
#endif
	
	/* Add loopback */
	config_loopback();

	/* Restore defaults if necessary */
	restore_defaults();

	/* Convert deprecated variables */
	convert_deprecated();

	/* Upgrade NVRAM variables to MBSS mode */
	upgrade_defaults();

    /* Read ethernet MAC, etc */
    eval("read_bd");

    /* Reset some wps-related parameters */
    nvram_set("wps_start",   "none");

    //qos_config_switch();

	/* Basic initialization */
	sysinit();

	/* Setup signal handlers */
	signal_init();
	signal(SIGHUP, rc_signal);
	signal(SIGUSR2, rc_signal);
	signal(SIGINT, rc_signal);
	signal(SIGALRM, rc_signal);
	signal(SIGUSR1, rc_signal);	
	signal(SIGQUIT, rc_signal); 
    signal(SIGILL,  rc_signal); 
	sigemptyset(&sigset);

	/* Give user a chance to run a shell before bringing up the rest of the system */
	if (!noconsole)
		run_shell(1, 0);

	/* Get boardflags to see if VLAN is supported */
#ifdef __CONFIG_VLAN__
	boardflags = strtoul(nvram_safe_get("boardflags"), NULL, 0);
#endif	/* __CONFIG_VLAN__ */


#if 0 
	/* Add loopback */
	config_loopback();

	/* Convert deprecated variables */
	convert_deprecated();

	/* ses cleanup of variables left half way through */
	ses_cleanup();

	/* ses_cl cleanup of variables left half way through */
	ses_cl_cleanup();

	/* Upgrade NVRAM variables to MBSS mode */
	upgrade_defaults();

	/* Restore defaults if necessary */
	restore_defaults();

    /* Read board data again, since the "restore_defaults" action
     * above will overwrite some of our settings */
    eval("read_bd");
#endif /* 0 */
    
#ifdef __CONFIG_NAT__
	/* Auto Bridge if neccessary */
	if (!strcmp(nvram_safe_get("auto_bridge"), "1"))
	{
		auto_bridge();
	}
	/* Setup wan0 variables if necessary */
	set_wan0_vars();
#endif	/* __CONFIG_NAT__ */
    /* create the USB semaphores */
#ifdef SAMBA_ENABLE
    usb_sem_init();
#endif

#ifdef KERNEL_LOG
	system("cat /proc/kmsg > /tmp/kernel_log.txt &");
#endif

	/* Loop forever */
	for (;;) {
		switch (state) {
		case RESTART:
			dprintf("RESTART\n");

            printf("---> rc restarted\n");
			/* Fall through */
            /* When vista finished configuring this router (wl0_wps_config_state: 0->1),
             * then we come here to restart WLAN 
             */
            stop_wps();
			stop_nas();
            stop_eapd();
			stop_bcmupnp();
			stop_wlan();

    	    convert_wlan_params();  
            sleep(2);               /* Wait some time for wsc, etc to terminate */

            /* if "unconfig" to "config" mode, force it to built-in registrar and proxy mode */
            printf("wps_status:%s\n", nvram_get("wps_status"));
            printf("lan_wps_oob:%s\n", nvram_get("lan_wps_oob"));

            if ( nvram_match("wps_status", "0") ) //restart wlan for wsc
            {
                nvram_set("wl_wps_reg", "enabled");
                nvram_set("wl0_wps_reg", "enabled");
#ifdef WLAN_REPEATER 
                if(!strcmp(nvram_get("lan_wps_oob"), "disabled"))
                {
                    nvram_set("wl0_wps_config_state","1");
                }else{
                    nvram_set("wl0_wps_config_state","0");
                }
#endif

                printf("restart -- wl0_wps_config_state=%s\n", nvram_get("wl0_wps_config_state"));
                nvram_set("wl_wps_config_state", nvram_get("wl0_wps_config_state"));
            }
            
            printf("wl0_wps_config_state:%s\n",
                    nvram_get("wl0_wps_config_state"));

            /* hide unnecessary warnings (Invaid XXX, out of range xxx etc...)*/
            {
                #include <fcntl.h>
                int fd1, fd2;
                fd1 = dup(2);
                fd2 = open("/dev/null", O_WRONLY);
                close(2);
                dup2(fd2, 2);
                close(fd2);
                start_wlan(); //<-- to hide messages generated here
                close(2);
                dup2(fd1, 2);
                close(fd1);
            }
            
            save_wlan_time();          
            start_bcmupnp();
            start_eapd();           
            start_nas();            
            start_wps();            
            sleep(2);               /* Wait for WSC to start */
            /* Must call it when start wireless */
            start_wl();
            nvram_commit();         /* Save WCN obtained parameters */

			state = IDLE;
#if 0       
			if (nvram_match("wla_secu_type", "None"))
			{
#if (defined BCM5356)
				system("/sbin/gpio 1 0");
#endif
			}
			else
			{
#if (defined BCM5356)
				system("/sbin/gpio 1 1");
#endif
			}
#endif      
            printf("---> rc stopped\n");
			break;

		case STOP:
			dprintf("STOP\n");
            
#if 1
            stop_wps();
            stop_nas();
            stop_eapd(); 
            stop_bcmupnp();
#else
			stop_services();
#ifdef __CONFIG_NAT__
			stop_wan();
#endif	/* __CONFIG_NAT__ */
#endif
			stop_lan();
#ifdef __CONFIG_VLAN__
			if (boardflags & BFL_ENETVLAN)
				stop_vlan();
#endif	/* __CONFIG_VLAN__ */
			if (state == STOP) {
				state = IDLE;
				break;
			}
			/* Fall through */
		case START:
			dprintf("START\n");
            
#ifdef WLAN_REPEATER
            nvram_set("gpio15", "wps_button");
            /* trying to connect rootAP using the highest speed.
             * which means "wla_mode" = "300Mbps" */
            nvram_set("wl0_nbw_cap", "1");
            nvram_set("wl0_nctrlsb", "lower");
#endif
			{ /* Set log level on restart */
				char *loglevel;
				int loglev = 8;

				if ((loglevel = nvram_get("console_loglevel"))) {
					loglev = atoi(loglevel);
				}
				klogctl(8, NULL, loglev);
				if (loglev < 7) {
					printf("WARNING: console log level set to %d\n", loglev);
				}
			}

			set_timezone();
#ifdef __CONFIG_VLAN__
			if (boardflags & BFL_ENETVLAN)
				start_vlan();
#endif	/* __CONFIG_VLAN__ */
            /* hide unnecessary warnings (Invaid XXX, out of range xxx etc...)*/
            {
                #include <fcntl.h>
                int fd1, fd2;
                fd1 = dup(2);
                fd2 = open("/dev/null", O_WRONLY);
                close(2);
                dup2(fd2, 2);
                close(fd2);
			    start_lan(); //<-- to hide messages generated here
                close(2);
                dup2(fd1, 2);
                close(fd1);
            }
			if (nvram_match("wla_repeater", "1")) {
                /* if repeater mode, del vlan1 from br0 and disable vlan */
                //system("/usr/sbin/brctl delif br0 vlan1");
                //system("/usr/sbin/et robowr 0x34 0x00 0x00");
            }


#ifdef WLAN_REPEATER
            eval("gpio", "0", "0"); /* Turn off LED of Link. */
#endif

            
#if 1
            save_wlan_time();
			start_bcmupnp();
            start_eapd();
            start_nas();
            sleep(1);
            start_wps();
            sleep(2);
#else
			start_services();
#ifdef __CONFIG_NAT__
			start_wan();
#endif	/* __CONFIG_NAT__ */
#ifdef __CONFIG_USBAP__
#endif
#endif
            start_wl();
            /* Now start ACOS services */
            eval("acos_init");
            eval("acos_service", "start");

            /* Start wsc if it is in 'unconfiged' state, and if PIN is not disabled */
            if (nvram_match("wl0_wps_config_state", "0") && !nvram_match("wsc_pin_disable", "1"))
            {
                /* if "unconfig" to "config" mode, force it to built-in registrar and proxy mode */
                nvram_set("wl_wps_reg", "enabled");
                nvram_set("wl0_wps_reg", "enabled");
                nvram_set("wps_proc_status", "0");
                nvram_set("wps_method", "1");
                //nvram_set("wps_config_command", "1");
            }

            //eval("wl", "interference", "2");    

			/* Fall through */
		case TIMER:
#if 0
			dprintf("TIMER\n");
			do_timer();
#endif
			/* Fall through */
		case IDLE:
			dprintf("IDLE\n");
			state = IDLE;

			if (nvram_match("wps_start",   "none"))
				send_wps_led_cmd(WPS_LED_BLINK_OFF, 0);

#ifdef CAPI_AP
			if (start_aput == TRUE) {
				system("/usr/sbin/wfa_aput_all&");
				start_aput = FALSE;
			}
#endif /* CAPI_AP */
			/* Wait for user input or state change */
			while (signalled == -1) {
				if (!noconsole && (!shell_pid || kill(shell_pid, 0) != 0))
					shell_pid = run_shell(0, 1);
				else
					sigsuspend(&sigset);
#ifdef __CONFIG_SHRINK_MEMORY__
				eval("cat", "/proc/shrinkmem");
#endif
			}
			state = signalled;
			signalled = -1;
			break;

		case WSC_RESTART:
			dprintf("WSC_RESTART\n");
			state = IDLE;
			stop_wps();    
			start_wps();    
			break;

#ifdef WLAN_REPEATER
        case WLAN_SLOW_START:
            /* In disconnecting with rootap, we hope station get IP from DUT,
             * so we make a connection break with station and slow start wifi.
             */
            system("/usr/sbin/et phywr 0 0x1e 0x80a8"); /* disable switch for 5357. */
#endif

            /* We come here only if user press "apply" in Wireless GUI */
		case WLANRESTART:
		    printf("--> rc wlanrestart, started\n");
            printf("%s\n", nvram_get("wps_device_pin"));
		    stop_wps(); 
		    stop_nas();
            stop_eapd();
            stop_bcmupnp();
			stop_wlan();
#if 0            
            printf("======> before read_bd.\n");
            printf("-> wps_randomssid:%s\n", nvram_get("wps_randomssid"));
            printf("-> wla_ssid_2:%s\n",    nvram_get("wla_ssid_2"));
            printf("-> wl0.1_ssid:%s\n",    nvram_get("wl0.1_ssid"));
            printf("-> wl0.1_akm:%s\n",     nvram_get("wl0.1_akm"));
            printf("-> wl0.1_crypto:%s\n",  nvram_get("wl0.1_crypto"));
            printf("-> wl0.1_wpa_psk:%s\n", nvram_get("wl0.1_wpa_psk"));
            printf("-> wl0_ssid:%s\n",    nvram_get("wl0_ssid"));
            printf("-> wl0_akm:%s\n",     nvram_get("wl0_akm"));
            printf("-> wl0_crypto:%s\n",  nvram_get("wl0_crypto"));
            printf("-> wl0_wpa_psk:%s\n", nvram_get("wl0_wpa_psk"));
            printf("--->> read_bd started.\n");
#endif
#ifdef WLAN_REPEATER
            printf("->%d\n", nvram_match("wla_secu_type_2", "WPA-PSK"));
            printf("->%d\n", nvram_match("wla_secu_type_2", "WEP"));

            if(nvram_match("wla_secu_type_2", "WPA-PSK") ||
                nvram_match("wla_secu_type_2", "WEP"))
                nvram_set("wla_mode", "g and b");
            else
                nvram_set("wla_mode", "300Mbps");
            printf("->%s\n", nvram_get("wla_mode"));
#endif
#if 0                       
            printf("<====== after read_bd.\n");
            printf("-> wps_randomssid:%s\n", nvram_get("wps_randomssid"));
            printf("-> wla_ssid_2:%s\n",    nvram_get("wla_ssid_2"));
            printf("-> wl0.1_ssid:%s\n",    nvram_get("wl0.1_ssid"));
            printf("-> wl0.1_akm:%s\n",     nvram_get("wl0.1_akm"));
            printf("-> wl0.1_crypto:%s\n",  nvram_get("wl0.1_crypto"));
            printf("-> wl0.1_wpa_psk:%s\n", nvram_get("wl0.1_wpa_psk"));
            printf("-> wl0_ssid:%s\n",    nvram_get("wl0_ssid"));
            printf("-> wl0_akm:%s\n",     nvram_get("wl0_akm"));
            printf("-> wl0_crypto:%s\n",  nvram_get("wl0_crypto"));
            printf("-> wl0_wpa_psk:%s\n", nvram_get("wl0_wpa_psk"));
#endif            
#ifdef WLAN_REPEATER
            #define waiting_sec 10
            if(state == WLAN_SLOW_START)
            {
                int i;
                for(i=0; i<waiting_sec; i++){
                    sleep(1);
                    printf("%d\n", i);
                }
                system("/usr/sbin/et phywr 0 0x1e 0x80a0"); /*enable switch for 5357. */
            }
#endif

            eval("read_bd");    
            nvram_set("wlanrestart_now", "1");       
            /* hide unnecessary warnings (Invaid XXX, out of range xxx etc...)*/
            {
                #include <fcntl.h>
                int fd1, fd2;
                fd1 = dup(2);
                fd2 = open("/dev/null", O_WRONLY);
                close(2);
                dup2(fd2, 2);
                close(fd2);
                start_wlan(); //<-- to hide messages generated here
                close(2);
                dup2(fd1, 2);
                close(fd1);
            }
            //printf("1.%s\n", nvram_get("wps_device_pin"));
            //sleep(2);
            save_wlan_time();
            start_bcmupnp();
            start_eapd();
            start_nas();
            //printf("2.%s\n", nvram_get("wps_device_pin"));
            //sleep(2);
            start_wps();
            //sleep(2);
            //printf("3.%s\n", nvram_get("wps_device_pin"));
            //sleep(2);           /* Wait for WSC to start */
            start_wl();
            printf("--> rc wlanrestart, stopped\n");
            printf("%s\n", nvram_get("wps_device_pin"));
            nvram_set("wlanrestart_now", "0");        
            /* Start wsc if it is in 'unconfiged' state */
            if (nvram_match("wl0_wps_config_state", "0") && !nvram_match("wsc_pin_disable", "1"))
            {
                /* if "unconfig" to "config" mode, force it to built-in registrar and proxy mode */
                nvram_set("wl_wps_reg", "enabled");
                nvram_set("wl0_wps_reg", "enabled");
                nvram_set("wps_proc_status", "0");
                nvram_set("wps_method", "1");
                //nvram_set("wps_config_command", "1");
            }
            
            if (nvram_match("wla_wlanstate","Enable"))
                eval("gpio", "23", "1", "1"); /* turn on wireless LED*/
            if (nvram_match("wla_wlanstate","Disable"))
                eval("gpio", "23", "0", "1"); /* turn off wireless LED */
            
			state = IDLE;

		    break;
		    
		default:
			dprintf("UNKNOWN\n");
			return;
		}
	}
}

int
main(int argc, char **argv)
{
	char *base = strrchr(argv[0], '/');

	base = base ? base + 1 : argv[0];

	/* init */
#ifdef LINUX26
	if (strstr(base, "preinit")) {
		mount("devfs", "/dev", "tmpfs", MS_MGC_VAL, NULL);
		mknod("/dev/console", S_IRWXU|S_IFCHR, makedev(5, 1));
#else /* LINUX26 */
	if (strstr(base, "init")) {
#endif /* LINUX26 */
		main_loop();
		return 0;
	}

	/* Set TZ for all rc programs */
	setenv("TZ", nvram_safe_get("time_zone"), 1);

	/* rc [stop|start|restart ] */
	if (strstr(base, "rc")) {
		if (argv[1]) {
			if (strncmp(argv[1], "start", 5) == 0)
				return kill(1, SIGUSR2);
			else if (strncmp(argv[1], "stop", 4) == 0)
				return kill(1, SIGINT);
			else if (strncmp(argv[1], "restart", 7) == 0)
				return kill(1, SIGHUP);
		    else if (strcmp(argv[1], "wlanrestart") == 0)
		        return kill(1, SIGQUIT);
#ifdef WLAN_REPEATER
            else if(strcmp(argv[1], "wlan_slow_start") == 0)
                return kill(1, SIGILL);
#endif
		} else {
			fprintf(stderr, "usage: rc\
                    [start|stop|restart|wlanrestart]|wlan_slow_start]\n");
			return EINVAL;
		}
	}

#ifdef __CONFIG_NAT__
	/* ppp */
	else if (strstr(base, "ip-up"))
		return ipup_main(argc, argv);
	else if (strstr(base, "ip-down"))
		return ipdown_main(argc, argv);

	/* udhcpc [ deconfig bound renew ] */
	else if (strstr(base, "udhcpc"))
		return udhcpc_wan(argc, argv);
#endif	/* __CONFIG_NAT__ */

	/* ldhclnt [ deconfig bound renew ] */
	else if (strstr(base, "ldhclnt"))
		return udhcpc_lan(argc, argv);

	/* stats [ url ] */
	else if (strstr(base, "stats"))
		return http_stats(argv[1] ? : nvram_safe_get("stats_server"));

	/* erase [device] */
	else if (strstr(base, "erase")) {
		/*
		if (argv[1] && ((!strcmp(argv[1], "boot")) ||
			(!strcmp(argv[1], "linux")) ||
			(!strcmp(argv[1], "rootfs")) ||
			(!strcmp(argv[1], "nvram")))) {
		*/
		if (argv[1]) {
			return mtd_erase(argv[1]);
		} else {
			fprintf(stderr, "usage: erase [device]\n");
			return EINVAL;
		}
	}

	/* write [path] [device] */
	else if (strstr(base, "write")) {
		if (argc >= 3)
			return mtd_write(argv[1], argv[2]);
		else {
			fprintf(stderr, "usage: write [path] [device]\n");
			return EINVAL;
		}
	}

	/* hotplug [event] */
	else if (strstr(base, "hotplug")) {
		if (argc >= 2) {
			if (!strcmp(argv[1], "net"))
				return hotplug_net();
#if (defined SAMBA_ENABLE || defined HSDPA) 
			else if (!strcmp(argv[1], "usb"))
				return hotplug_usb();
#endif
#if 0
			else if (!strcmp(argv[1], "block"))
                                return hotplug_block();
#endif
		} else {
			fprintf(stderr, "usage: hotplug [event]\n");
			return EINVAL;
		}
	}

	return EINVAL;
}
