/***************************************************************************
***
***    Copyright 2005  Hon Hai Precision Ind. Co. Ltd.
***    All Rights Reserved.
***    No portions of this material shall be reproduced in any form without the
***    written permission of Hon Hai Precision Ind. Co. Ltd.
***
***    All information contained in this document is Hon Hai Precision Ind.  
***    Co. Ltd. company private, proprietary, and trade secret property and 
***    are protected by international intellectual property laws and treaties.
***
****************************************************************************/
/****************************************************************************
 *
 *              ap51_main.c 
 *              ==============
 *
 *
 * 
 *  Author:     Jamsine Yang
 *  Date:       October, 2005
 *  Description:
 *                 Packet kernel image and rootfs image tools
 *
 *  History:
 *
 *  Modify Reason                                             Author            Date              Search Flag(Option)
 *-----------------------------------------------------------------------------------
 *  
 ***************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ap51_packet.h"

#define default_kernelname "ap71-kernel"
#define default_rootfsname "ap71-squashfs"
#define ap51_kernel_rootfs_image "ap71_kernel_rootfs_image.chk"
#define ap51_kernel_image "ap71_kernel_image.chk"
#define ap51_rootfs_image "ap71_rootfs_image.chk"

#define kernel_rootfs_image "_kernel_rootfs_image.chk"
#define kernel_image "_kernel_image.chk"
#define rootfs_image "_rootfs_image.chk"
#define image_chk ".chk"

#define install_path "./"       //"../image/"
#define boardIdfile "../tools/packet/compatible.txt"
#define MAX_HEADER_FILE_LINES 200
char rootfsname[MAX_FILENAME_LEN];
char kernelname[MAX_FILENAME_LEN];
char board_id[MAX_SIZE_BOARDID];
char boardIdname[MAX_FILENAME_LEN];
char out_kernel_rootfs_name[MAX_FILENAME_LEN];
char out_kernel_name[MAX_FILENAME_LEN];
char out_rootfs_name[MAX_FILENAME_LEN];
char region[MAX_FILENAME_LEN];
char version[MAX_FILENAME_LEN];
char cfg_file_name[MAX_FILENAME_LEN];
unsigned int regionIndex = 0;
char *Region_List[] = {
    "NONE_VERSION",             /* 0x30 */
    "WW_VERSION",               /* 0x31 */
    "NA_VERSION",               /* 0x32 */
    "JP_VERSION",               /* 0x33 */
    "GR_VERSION",               /* 0x34 */
    "PR_VERSION",               /* 0x35 */
    "KO_VERSION",               /* 0x36 */
    "RU_VERSION",               /* 0x37 */
    "SS_VERSION",               /* 0x38 */
    NULL
};
static void usage (void)
{
    printf("Version: 2.0\n");
    fprintf (stderr,
             "usage: packet -k [kernel file name] -f [rootfs file name] -b [board id file name]"
             " -ok [output kernel file name] -oall [output kernel and rootfs file name] -or [output rootfs file name]"
             " -i [configure file path/name]\r\n" "Example:\r\n"
             //"packet -k linux.trx -f rootfs -ok kernel -oall kernel_rootfs -or rootfs -r NA_VERSION -v V0.1.0_1.0.1\r\n");
             "packet -k linux.trx -f rootfs -b compatible.txt -ok kernel -oall kernel_rootfs -or rootfs -i ../../../project/acos/include/ambitCfg.h\r\n");
    exit (0);
}

int main (int argc, char **argv)
{
    unsigned char *buffer = (unsigned char *) NULL;
    unsigned char *kernelbuf = (unsigned char *) NULL;
    unsigned char *roootfsbuf = (unsigned char *) NULL;
    unsigned char *cfgbuf = (unsigned char *) NULL;
    FILE *kernelfileFp = (FILE *) NULL;
    FILE *roootfsfileFp = (FILE *) NULL;
    FILE *boardidfileFp = (FILE *) NULL;
    FILE *cfgFp = (FILE *) NULL;
    FILE *newfile0Fp = (FILE *) NULL;
    FILE *newfile1Fp = (FILE *) NULL;
    FILE *newfile2Fp = (FILE *) NULL;
    unsigned char *read_data = (unsigned char *) NULL;
    int roootfslen = 0, kernellen = 0, totalLen, totalLines = 0;
    int ret = -1, i, index = 0;
    unsigned char pData0 = 0;
    int dataPos;
    unsigned long fileSize = 897840, headerlen;
    ImageInfo_s image;
    unsigned char *roootfslenPtr = (unsigned char *) NULL, *headerPtr =
        (unsigned char *) NULL;
    unsigned long kernel_len = 0, rootfs_len = 0, image_chksum = 0;
    char region_tmp[MAX_FILENAME_LEN];
    char version_tmp[MAX_FILENAME_LEN];
    BOOL hasFound = FALSE;
    char *regionPtr = NULL, *versionSW = NULL, *versionUI = NULL, *tmpPtr =
        NULL;

    /* Skip program name */
    --argc;
    ++argv;

    if (!*argv)
        usage ();

    memset (board_id, '\0', MAX_SIZE_BOARDID);
    memset (rootfsname, '\0', MAX_FILENAME_LEN);
    memset (kernelname, '\0', MAX_FILENAME_LEN);
    memset (boardIdname, '\0', MAX_FILENAME_LEN);
    memset (out_kernel_rootfs_name, '\0', MAX_FILENAME_LEN);
    memset (out_kernel_name, '\0', MAX_FILENAME_LEN);
    memset (out_rootfs_name, '\0', MAX_FILENAME_LEN);
    memset (region, '\0', MAX_FILENAME_LEN);
    memset (version, '\0', MAX_FILENAME_LEN);
    memset (cfg_file_name, '\0', MAX_FILENAME_LEN);
    memset (region_tmp, '\0', MAX_FILENAME_LEN);
    memset (version_tmp, '\0', MAX_FILENAME_LEN);

    strcpy (rootfsname, install_path);
    strcpy (kernelname, install_path);


/* Process the remaining arguments. */
    for (; *argv; argv++)
    {
        if (!strncmp (*argv, "-k", 2))
        {
            if (*++argv)
                strcat (kernelname, *argv);
        }
        else if (!strncmp (*argv, "-f", 2))
        {
            if (*++argv)
                strcat (rootfsname, *argv);
        }
        else if (!strncmp (*argv, "-b", 2))
        {
            if (*++argv)
                strcat (boardIdname, *argv);
        }
        else if (!strncmp (*argv, "-ok", 3))
        {
            if (*++argv)
            {
                strncpy (out_kernel_name, *argv,
                         MAX_FILENAME_LEN - strlen (kernel_image) - 1);
                strcat (out_kernel_name, image_chk);
            }
        }
        else if (!strncmp (*argv, "-oall", 5))
        {
            if (*++argv)
            {
                strncpy (out_kernel_rootfs_name, *argv,
                         MAX_FILENAME_LEN - strlen (kernel_rootfs_image) - 1);
                strcat (out_kernel_rootfs_name, image_chk);
            }
        }
        else if (!strncmp (*argv, "-or", 3))
        {
            if (*++argv)
            {
                strncpy (out_rootfs_name, *argv,
                         MAX_FILENAME_LEN - strlen (rootfs_image) - 1);
                strcat (out_rootfs_name, image_chk);
            }
        }
        else if (!strncmp (*argv, "-r", 2))
        {
            if (*++argv)
            {
                strcpy (region, *argv);
            }
        }
        else if (!strncmp (*argv, "-v", 2))
        {
            if (*++argv)
                strcpy (version, *argv);
        }
        else if (!strncmp (*argv, "-i", 2))
        {
            if (*++argv)
                strcpy (cfg_file_name, *argv);
        }
        else
            usage ();

        if (!*argv)
            break;
    }

    if (!strlen (kernelname))
        strcpy (kernelname, default_kernelname);
    if (!strlen (rootfsname))
        strcpy (rootfsname, default_rootfsname);

    if (!strlen (out_kernel_rootfs_name))
        strcpy (out_kernel_rootfs_name, kernel_rootfs_image);
    if (!strlen (out_kernel_name))
        strcpy (out_kernel_name, kernel_image);
    if (!strlen (out_rootfs_name))
        strcpy (out_rootfs_name, rootfs_image);
    if (!strlen (boardIdname))
        strcpy (boardIdname, boardIdfile);

    if ((buffer = (unsigned char *) malloc (MAX_SIZE_FILE)) == NULL)
    {
        printf ("No memory buffer for using\n");
        goto end;
    }
    memset (buffer, '\0', MAX_SIZE_FILE);
    if ((kernelbuf = (unsigned char *) malloc (MAX_SIZE_KERNEL)) == NULL)
    {
        printf ("No memory buffer for using\n");
        goto end;
    }
    memset (kernelbuf, '\0', MAX_SIZE_KERNEL);
    if ((roootfsbuf = (unsigned char *) malloc (MAX_SIZE_FILE)) == NULL)
    {
        printf ("No memory buffer for using\n");
        goto end;
    }
    memset (roootfsbuf, '\0', MAX_SIZE_FILE);
    if ((cfgbuf = (unsigned char *) malloc (READ_BUFSZ)) == NULL)
    {
        printf ("No memory buffer for using\n");
        goto end;
    }
    memset (cfgbuf, '\0', READ_BUFSZ);
    if ((kernelfileFp = fopen (kernelname, "rb")) == NULL)
    {
        printf ("Can't open %s file.\n", kernelname);
        goto end;
    }
    if ((roootfsfileFp = fopen (rootfsname, "rb")) == NULL)
    {
        printf ("Can't open %s file.\n", rootfsname);
        goto end;
    }
    if ((boardidfileFp = fopen (boardIdname, "rb")) == NULL)
    {
        printf ("Can't open %s file.\n", boardIdname);
        goto end;
    }
    if ((newfile0Fp = fopen (out_kernel_rootfs_name, "wb+")) == NULL)
    {
        printf ("Can't open %s file.\n", out_kernel_rootfs_name);
        goto end;
    }
    if ((newfile1Fp = fopen (out_kernel_name, "wb+")) == NULL)
    {
        printf ("Can't open %s file.\n", out_kernel_name);
        goto end;
    }
    if ((newfile2Fp = fopen (out_rootfs_name, "wb+")) == NULL)
    {
        printf ("Can't open %s file.\n", out_rootfs_name);
        goto end;
    }
    if (strlen (cfg_file_name))
        if ((cfgFp = fopen (cfg_file_name, "rb")) == NULL)
        {
            printf ("Can't open %s file.\n", cfg_file_name);
            goto end;
        }
    kernellen =
        fread ((unsigned char *) kernelbuf, sizeof (char), MAX_SIZE_KERNEL,
               kernelfileFp);


    roootfslen =
        fread ((unsigned char *) roootfsbuf, sizeof (char), MAX_SIZE_FILE,
               roootfsfileFp);

    /* Foxconn added start Peter Ling 02/27/2006 */
    /* Pad rootfs to multiple of 65536 bytes + 1 byte.
     * The last byte will not write to flash by AP71 (for unknown reason).
     */
#if 0
    if ((roootfslen % 0x10000) != 0)
    {
        int paddedLen;

        paddedLen = roootfslen / 65536;
        paddedLen += 1;
        paddedLen *= 65536;
        paddedLen += 1;         /* To fake AP71 that we have 1 more sector */
        printf ("new rootfs len = 0x%lx (old = 0x%lx)\n", paddedLen,
                roootfslen);
        roootfslen = paddedLen;
    }
    else
    {
        roootfslen += 1;
        printf ("new rootfs len = 0x%lx\n", roootfslen);
    }
#endif
    /* Foxconn added end Peter Ling 02/27/2006 */

    fread ((unsigned char *) board_id, sizeof (char), MAX_SIZE_BOARDID,
           boardidfileFp);

    if (board_id[strlen (board_id) - 1] == 0x0A)
        board_id[strlen (board_id) - 1] = '\0';

    /* printf ("boardid = '%s'\n", board_id); */

    if (kernellen > MAX_SIZE_KERNEL)
    {
        printf ("The kernel image size is over %dMbytes!!\n",
                MAX_SIZE_KERNEL / (1024 * 1024));
        goto end;
    }

    if (roootfslen > MAX_SIZE_FILE)
    {
        printf ("The rootfs image size is over %dMbytes!!\n",
                MAX_SIZE_FILE / (1024 * 1024));
        goto end;
    }
    /* Parse Region and Version of firmware */
    do
    {
        if (fgets ((unsigned char *) cfgbuf, READ_BUFSZ - 1, cfgFp) == NULL)
        {
            break;
        }
        if (strncmp (cfgbuf, "#define", strlen ("#define")) != 0)
        {
            totalLines++;
            continue;
        }

        tmpPtr = strstr (cfgbuf, "\r\n");
        if (tmpPtr)
            *tmpPtr = '\0';
        if ((tmpPtr = strstr (cfgbuf, " FW_REGION ")) != NULL)
        {
            regionPtr = tmpPtr + strlen (" FW_REGION ");
            index = 0;
            while (Region_List[index] != NULL)
            {
                if (strstr (regionPtr, Region_List[index]) != 0)
                {
                    regionIndex = index;
                    //printf ("regionIndex = 0x%x\r\n", regionIndex);
                    break;
                }
                index++;
            }
        }
        else if ((tmpPtr =
                  strstr (cfgbuf, " AMBIT_SOFTWARE_VERSION ")) != NULL)
        {
            versionSW = tmpPtr + strlen (" AMBIT_SOFTWARE_VERSION ");
            while (*versionSW != '\0')
            {
                if (*versionSW != ' ')
                {
                    strcpy (version, versionSW);
                    strcat (version, "_");
                    //printf ("version = %s\r\n", version);
                    break;
                }
                else
                    versionSW++;
            }
        }
        else if ((tmpPtr = strstr (cfgbuf, " AMBIT_UI_VERSION ")) != NULL)
        {
            versionUI = tmpPtr + strlen (" AMBIT_UI_VERSION ");
            while (*versionUI != '\0')
            {
                if (*versionUI != ' ')
                {
                    strcat (version, versionUI);
                    hasFound = TRUE;
                    //printf ("version = %s\r\n", version);
                    break;
                }
                else
                    versionUI++;
            }
        }
        totalLines++;
    }
    while (hasFound == FALSE && totalLines < MAX_HEADER_FILE_LINES);

    /* Packet rootfs & kernel image into a file */
    fileSize =
        addHeader (board_id, (char *) buffer, (char *) kernelbuf,
                   (char *) roootfsbuf, kernellen, roootfslen);

    if (fileSize > MAX_SIZE_FILE)
    {
        printf ("The image size is over %dMbytes!!\n",
                MAX_SIZE_FILE / (1024 * 1024));
        goto end;
    }

    totalLen = fwrite ((char *) buffer, sizeof (char), fileSize, newfile0Fp);
    memset (buffer, '\0', MAX_SIZE_FILE);
    /* Packet kernel image */
    fileSize =
        addHeader (board_id, (char *) buffer, (char *) kernelbuf,
                   (char *) roootfsbuf, kernellen, 0);

    if (fileSize > MAX_SIZE_KERNEL)
    {
        printf ("The image size is over %dMbytes!!\n",
                MAX_SIZE_KERNEL / (1024 * 1024));
        goto end;
    }

    totalLen = fwrite ((char *) buffer, sizeof (char), fileSize, newfile1Fp);
    memset (buffer, '\0', MAX_SIZE_FILE);
    /* Packet rootfs image */
    fileSize =
        addHeader (board_id, (char *) buffer, (char *) kernelbuf,
                   (char *) roootfsbuf, 0, roootfslen);

    if (fileSize > MAX_SIZE_FILE)
    {
        printf ("The image size is over %dMbytes!!\n",
                MAX_SIZE_FILE / (1024 * 1024));
        goto end;
    }

    totalLen = fwrite ((char *) buffer, sizeof (char), fileSize, newfile2Fp);
    goto end;

  end:
    if (kernelfileFp)
        fclose (kernelfileFp);

    if (roootfsfileFp)
        fclose (roootfsfileFp);

    if (newfile0Fp)
        fclose (newfile0Fp);

    if (newfile1Fp)
        fclose (newfile1Fp);

    if (newfile2Fp)
        fclose (newfile2Fp);

    if (buffer)
        free (buffer);

    if (roootfsbuf)
        free (roootfsbuf);

    if (kernelbuf)
        free (kernelbuf);

    if (cfgbuf)
        free (cfgbuf);
    return 0;
}
