// stdafx.cpp : source file that includes just the standard includes
//      ap51.pch will be the pre-compiled header
//      stdafx.obj will contain the pre-compiled type information

#include "ap51_packet.h"

//#include <features.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
extern char region[];
extern char version[];
extern unsigned int regionIndex;


typedef struct MemBlock_s       /* Memory block structure */
{
    unsigned char **pData;      /* pointer to a 2-D array (pData[m][n]) to store the downloaded image */
    unsigned long numOfBlock;   /* == m */
    unsigned long sizeOfBlock;  /* == n */
    unsigned long blockIdx;     /* index of the block being accessed ( 0 <= blockIdx < m) */
    unsigned long dataIdx;      /* index within the block being accessed ( 0 <= dataIdx < n) */
    unsigned long fileSize;     /* Stores the total size (in bytes) of downloaded image, including the file header */
    ImageInfo_s image;          /*Add by Jasmine, 11/18/2005 */
} MemBlock_t;

static MemBlock_t memBlock;

static unsigned long calculate_checksum (int action, char *s, int size)
{
    static UINT32 c0, c1;
    UINT32 checksum, b;
    int i;

    switch (action)
    {
    case 0:
        c0 = c1 = 0;
        break;
    case 1:
        for (i = 0; i < size; i++)
        {
            c0 += s[i] & 0xff;
            c1 += c0;
        }
        break;
    case 2:
        b = (c0 & 65535) + ((c0 >> 16) & 65535);
        c0 = ((b >> 16) + b) & 65535;

        b = (c1 & 65535) + ((c1 >> 16) & 65535);
        c1 = ((b >> 16) + b) & 65535;

        checksum = ((c1 << 16) | c0);

        return checksum;
    }
    return TRUE;
}


unsigned long addHeader (char *board_id, char *bufPtr, char *kernelPtr,
                         char *rootfsPtr, UINT32 kernelSize,
                         UINT32 rootfsSize)
{
    unsigned long headerChecksum = 0, imageChecksum = 0, kernelChecksum =
        0, rootfsChecksum = 0;
    unsigned long headeroffset = 0, totalHeaderLen = 0, boardIDLen =
        0, imageLen = 0;
    imageHdr_s header;
    unsigned char *buffer = (unsigned char *) NULL, *imageBuf =
        (unsigned char *) NULL;
    char boardIDbuf[MAX_SIZE_BOARDID];
    unsigned long i = 0, tmpLen = 0, k = 0, index = 0;
    unsigned int aa, bb, cc, dd, ee, ff;
    int ret = 0; /* wklin added, 10/03/2008 */
    unsigned int gg = 0; /* wklin added, 10/03/2008 */

    if ((buffer =
         (unsigned char *) malloc (sizeof (imageHdr_s) + MAX_SIZE_BOARDID)) ==
        (unsigned char *) NULL)
    {
        printf ("No memory buffer for using\n");
        goto endline;
    }

    if ((imageBuf =
         (unsigned char *) malloc (kernelSize + rootfsSize + 300)) ==
        (unsigned char *) NULL)
    {
        printf ("No memory buffer for using\n");
        goto endline;
    }
    if (rootfsSize)
        memcpy (imageBuf, rootfsPtr, rootfsSize);

    if (kernelSize)
        memcpy (&imageBuf[rootfsSize], kernelPtr, kernelSize);

    /* foxconn added start, wenchia, 2006/03/20 */
    if (rootfsSize)
    {
        calculate_checksum (0, (char *) NULL, 0);
        calculate_checksum (1, (char *) rootfsPtr, rootfsSize);
        rootfsChecksum = calculate_checksum (2, (char *) NULL, 0);
    }

    if (kernelSize)
    {
        calculate_checksum (0, (char *) NULL, 0);
        calculate_checksum (1, (char *) kernelPtr, kernelSize);
        kernelChecksum = calculate_checksum (2, (char *) NULL, 0);
    }
    /* foxconn added end, wenchia, 2006/03/20 */

    imageLen = kernelSize + rootfsSize;
    calculate_checksum (0, (char *) NULL, 0);
    calculate_checksum (1, (char *) imageBuf, imageLen);
    imageChecksum = calculate_checksum (2, (char *) NULL, 0);

    /* build header */
    memset (boardIDbuf, '\0', MAX_SIZE_BOARDID);
    strcpy (boardIDbuf, board_id);

    //strcat(boardIDbuf,pipeChar);
    headeroffset = sizeof (imageHdr_s);
    boardIDLen = strlen (boardIDbuf);
    totalHeaderLen = headeroffset + boardIDLen;

    memcpy (header.magic, FIELD_AMBIT_IDENTIFY, FIELD_LEN_SIZE);
    for (i = 0; i < FIELD_LEN_SIZE; i++)
    {
        k = (3 - i) * 8;
        tmpLen = (totalHeaderLen >> k) & 0x000000ff;
        header.header_len[i] = tmpLen;
    }

    /* foxconn modified start, wenchia, 2006/03/20 */
    memset (header.reserved, '\0', 8);
    /* Jasmine Yang Modify start, 10/25/2006 */    
    header.reserved[0] = regionIndex;
    //strcpy(version, "V1.0.2_1.0.3");
    /* wklin modified start, 10/03/2008 */
    if(version[0] = '"')  /* e.g. "V1.0.2"_"1.0.3" */
        ret = sscanf (version, "\"V%d.%d.%d\"_\"%d.%d.%d\"", &aa, &bb, &cc, &dd, &ee, &ff);
    else  /* e.g. V1.0.2_1.0.3 */
        ret = sscanf (version, "V%d.%d.%d_%d.%d.%d", &aa, &bb, &cc, &dd, &ee, &ff);
    if (ret != 6) {
        /* may be 4+3 version */
        if(version[0] = '"')  /* e.g. "V1.0.2.0"_"1.0.3" */
            ret = sscanf (version, "\"V%d.%d.%d.%d\"_\"%d.%d.%d\"", &aa, &bb, &cc, &dd, &ee, &ff,&gg);
        else  /* e.g. V1.0.2.0_1.0.3 */
            ret = sscanf (version, "V%d.%d.%d.%d_%d.%d.%d", &aa, &bb, &cc, &dd, &ee, &ff,&gg);
    }
    //printf("\r\nversion=%s,aa=%d,bb=%d,cc=%d,dd=%d,ee=%d,ff=%d,gg=%d\r\n", version, aa,bb,cc,dd,ee,ff,gg);
    header.reserved[1] = aa;
    header.reserved[2] = bb;
    header.reserved[3] = cc;
    header.reserved[4] = dd;
    header.reserved[5] = ee;
    header.reserved[6] = ff;
    header.reserved[7] = gg;
    /* wklin modified end, 10/03/2008 */
    /* Jasmine Yang Modify end, 10/25/2006 */
    for (i = 0; i < FILES_CHKLEN_SIZE; i++)
    {
        k = (FILES_CHKLEN_SIZE - 1 - i) * 8;
        tmpLen = (rootfsChecksum >> k) & 0x000000ff;
        header.rootfs_chksum[i] = tmpLen;
    }
    for (i = 0; i < FILES_CHKLEN_SIZE; i++)
    {
        k = (FILES_CHKLEN_SIZE - 1 - i) * 8;
        tmpLen = (kernelChecksum >> k) & 0x000000ff;
        header.kernel_chksum[i] = tmpLen;
    }
    /* foxconn modified end, wenchia, 2006/03/20 */

    for (i = 0; i < FIELD_LEN_SIZE; i++)
    {
        k = (3 - i) * 8;
        if (kernelSize)
            tmpLen = (kernelSize >> k) & 0x000000ff;
        else
            tmpLen = 0x00;
        header.kernel_len[i] = tmpLen;
    }
    for (i = 0; i < 4; i++)
    {
        k = (3 - i) * 8;
        if (rootfsSize)
            tmpLen = (rootfsSize >> k) & 0x000000ff;
        else
            tmpLen = 0x00;
        header.rootfs_len[i] = tmpLen;
    }
    for (i = 0; i < FILES_CHKLEN_SIZE; i++)
    {
        k = (FILES_CHKLEN_SIZE - 1 - i) * 8;
        tmpLen = (imageChecksum >> k) & 0x000000ff;
        header.image_chksum[i] = tmpLen;
    }
    memset (header.header_chksum, '\0', FILES_CHKLEN_SIZE);
    memcpy (&buffer[0], (unsigned char *) &header, headeroffset);
    memcpy (&buffer[headeroffset], (char *) boardIDbuf, boardIDLen);
    calculate_checksum (0, (char *) NULL, 0);
    calculate_checksum (1, (char *) &buffer[0], headeroffset + boardIDLen);
    headerChecksum = calculate_checksum (2, (char *) NULL, 0);
    for (i = 0; i < FILES_CHKLEN_SIZE; i++)
    {
        k = (FILES_CHKLEN_SIZE - 1 - i) * 8;
        tmpLen = (headerChecksum >> k) & 0x000000ff;
        header.header_chksum[i] = tmpLen;
    }
    memcpy (&bufPtr[0], (unsigned char *) &header, headeroffset);
    memcpy (&bufPtr[headeroffset], (char *) &boardIDbuf[0], boardIDLen);

    memcpy (&bufPtr[totalHeaderLen], &imageBuf[0], imageLen);
    goto endline;

  endline:
    if (imageBuf)
        free (imageBuf);

    if (buffer)
        free (buffer);

    return (totalHeaderLen + imageLen);
}
